#!/usr/bin/env python
# -*- coding: utf-8


import logging
import os
from ..adapter import gen_examples as gen_example
from ..adapter import get_imp_cls_def as get_imp_cls
from ..adapter.utils.xml_parse.xml_parser_get_xmlns import get_xmlns_info, xmlns_info_without_prefix, \
    get_xml_namespace
from ..adapter.get_argument_spec_documentation import get_xml_options, get_xml_dict, get_xml_descption, \
    gen_documentation, check_xpaths_exists, get_xml_key_leafs, get_xml_list_leafs, get_xml_leafinfos, gen_argument_spec
from ..adapter.get_rpc_head_tail import get_rpc_head, get_rpc_tail

yang_handler = None

head = """#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

"""


def gen_documentation_params(full_xml_file_path, full_xml_ordered_dict, script_name, xml_namespace, xmlns_info):
    """
     get params of the xml_file
    :param full_xml_file_path:
    :param full_xml_ordered_dict:
    :param script_name:
    :return:
    """
    xml_options = get_xml_options(full_xml_file_path, full_xml_ordered_dict, xmlns_info)
    if len(full_xml_ordered_dict) == 0:
        logging.error("xml file %s can't be tranport to OrderedDict", full_xml_file_path)
        return

    if not check_xpaths_exists(full_xml_ordered_dict, xmlns_info):
        logging.error("not all node defined in xml %s occured in yang files", os.path.basename(full_xml_file_path))
        return
    module_desc = get_xml_descption(xml_namespace)  # Get the overall description of the message
    xml_type = get_head_content(full_xml_ordered_dict)[2]
    documentation = gen_documentation(module_desc, xml_options, script_name, xml_type)
    return documentation


def get_head_content(full_xml_ordered_dict):
    """Get xml head"""
    head_content = get_rpc_head(full_xml_ordered_dict)
    return head_content


def get_xml_tail(full_xml_ordered_dict):
    """Get xml tail"""
    xml_tail = get_rpc_tail(full_xml_ordered_dict)
    return xml_tail


def get_user_check_stmts_module(user_def_dir, filename):
    """Get user-check script's content."""
    if user_def_dir:
        user_def_file = os.path.join(os.path.join(user_def_dir, filename + ".py"))
    else:
        user_def_file = ''
    if os.path.isfile(user_def_file):
        logging.info("%s exist, get UserCheck definaiton and import statements from it", user_def_file)
        module = get_imp_cls.get_imp_stmts(user_def_file)
        user_check_stmts = get_imp_cls.get_user_check_cls_def(user_def_file)
    else:
        module = ''
        user_check_stmts = get_imp_cls.USER_CHECK_STMT
    return [module, user_check_stmts]


def get_params(full_xml_file_path, script_name, output_file_path, user_def_dir=None):
    """Get params"""
    full_xml_ordered_dict = get_xml_dict(full_xml_file_path)
    if full_xml_ordered_dict is None:
        return {}
    if len(full_xml_ordered_dict) == 0:
        return {}
    # Get type of xml.
    pkg_type = list(full_xml_ordered_dict.keys())[0]
    if full_xml_ordered_dict[pkg_type] is None:
        return {}
    # Get xml namesapce.
    xml_namespace = get_xml_namespace(full_xml_file_path,pkg_type)
    # Get xmlns_info.
    namespace = get_xmlns_info(full_xml_file_path,pkg_type)
    xmlns_info = xmlns_info_without_prefix(namespace)

    # Get documentation
    documentation = gen_documentation_params(full_xml_file_path, full_xml_ordered_dict, script_name, xml_namespace, xmlns_info)
    # Get key_list
    key_list = get_xml_key_leafs(full_xml_ordered_dict, xmlns_info)
    # Get list_list
    list_list = get_xml_list_leafs(full_xml_ordered_dict, xmlns_info)
    # Get leaf_info
    leaf_info = get_xml_leafinfos(full_xml_ordered_dict, xmlns_info)
    # Get argument_spec
    argument_spec = gen_argument_spec(full_xml_ordered_dict, xmlns_info, key_list, list_list)
    # Get example
    example = gen_example.create_example(full_xml_file_path, script_name, pkg_type, argument_spec)
    # Get head_content
    head_content = get_head_content(full_xml_ordered_dict)
    # Get xml_tail
    xml_tail = get_xml_tail(full_xml_ordered_dict)
    # Get module
    module = get_user_check_stmts_module(user_def_dir, script_name)[0]
    # Get user_check_stmts
    user_check_stmts = get_user_check_stmts_module(user_def_dir, script_name)[1]
    return {"head": head,
            "leaf_info": leaf_info,
            "argument_spec": argument_spec,
            "documentation": documentation,
            "example": example,
            "namespace": namespace,
            "head_content": head_content,
            "xml_tail": xml_tail,
            "output_file_path": output_file_path,
            "filename": script_name,
            "module": module,
            "user_check_stmts": user_check_stmts,
            "key_list": key_list}



#!/usr/bin/env python
# -*- coding: utf-8


import re
import os
import logging
from lxml import etree
from collections import OrderedDict
from xml.parsers.expat import ExpatError
import xmltodict
from ..base_util import error_write, xml_structure_except

COMMON_XMLNS_MAP = {"nc": "urn:ietf:params:xml:ns:netconf:base:1.0"}


def remove_namespace(file_name):
    """Get xml string without xmlns and prefix.
    Args:
        file_name: The path of xml file.
    Returns:
        remove_xmlns_str: The xml string without xmlns and prefix.
    Raises:
        Exception: Capture IO exception.
    """
    remove_xmlns_str = ''
    try:
        with open(file_name, 'r') as xml_file:
            xml_str = xml_file.read()
        if xml_str:
            attr_names = list()
            for item in re.findall(r'xmlns:(.*?)=".*?"', xml_str):
                attr_names.append(item)
            for item in re.findall(r"xmlns:(.*?)='.*?'", xml_str):
                attr_names.append(item)
            remove_xmlns_str = xml_str
            # if attr_name is not None:
            if attr_names:
                for attr_name in attr_names:
                    # remove: [xmlns:nc="xxx"] or [xmlns="xxx"]
                    remove_xmlns_str = re.sub(r'.xmlns(:' + attr_name + ')?=".*?"', '', remove_xmlns_str)
                    remove_xmlns_str = re.sub(r".xmlns(:" + attr_name + ")?='.*?'", "", remove_xmlns_str)
                for attr_name in attr_names:
                    # remove: [ nc: ] from <xxx nc:operation="merge">
                    remove_xmlns_str = re.sub(r'\s' + attr_name + ':', " ", remove_xmlns_str)
                    # remove: [ prefix: ] from <prefix:xxx>
                    remove_xmlns_str = re.sub(r'<' + attr_name + ':', "<", remove_xmlns_str)
                    # remove: [ prefix: ] from </prefix:xxx>
                    remove_xmlns_str = re.sub(r'</' + attr_name + ':', "</", remove_xmlns_str)
            else:
                # remove: [xmlns="xxx"]
                remove_xmlns_str = re.sub(r'.xmlns=".*?"', "", remove_xmlns_str)
                remove_xmlns_str = re.sub(r".xmlns='.*?'", "", remove_xmlns_str)
    except Exception as error_msg:
        error_write(str(error_msg))
    finally:
        return remove_xmlns_str

def xml_to_ordered_dict(file_name):
    """Parse the xml file to ordered dict.
    Args:
        file_name: The path of xml file.
    Returns:
        xml_structure_dict: An OrderedDict. Mapping the xml's structure.
        for example:
            xml_structure_dict:
                OrderedDict([('config',
                    OrderedDict([('interfaces',
                        OrderedDict([('interface',
                            OrderedDict([('name', None), ('ipv4',
                                OrderedDict([('address',
                                    OrderedDict([('prefix-length', None), ('ip', None)]))]))]))]))]))])
    Raises:
        Exception: Capture execution exception.
    """
    xml_structure_dict = OrderedDict()
    operation_name = os.path.abspath(os.path.join(file_name, os.pardir, os.pardir))
    input_action_match = operation_name.split(os.sep)[-1]
    xml_filtered_str = remove_namespace(file_name)
    if input_action_match == "rpcs":
        rpc_root = etree.fromstring(xml_filtered_str.encode('utf-8')).getroottree()
        root = rpc_root.xpath('//rpc')
        xml_type = root[0].tag
        root_string = etree.tostring(root[0])
        xml_dic = xmltodict.parse(root_string, xml_attribs=False)
        # xml_dic = xmltodict.parse(root_string)
        for value in xml_dic.values():
            temp_ordered_dict = value
        xml_structure_dict[xml_type] = temp_ordered_dict
        return xml_structure_dict
    if xml_filtered_str:
        # remove: [ type="subtree"  ] from <filter type="subtree">
        xml_filtered_str = re.sub(r'.type="(.*?)"', "", xml_filtered_str)
        xml_filtered_str = re.sub(r".type='(.*?)'", "", xml_filtered_str)
        # remove operation attribute.
        operations = list()
        for item in re.findall(r'.operation="(.*?)"', xml_filtered_str):
            operations.append(item)
        for item in re.findall(r".operation='(.*?)'", xml_filtered_str):
            operations.append(item)
        if operations:
            for operation in operations:
                xml_filtered_str = re.sub(r'.operation="(' + operation + ')"', "", xml_filtered_str)
                xml_filtered_str = re.sub(r".operation='(" + operation + ")'", "", xml_filtered_str)
        try:
            rpc_root = etree.fromstring(xml_filtered_str.encode('utf-8')).getroottree()
            root = rpc_root.xpath('//*/config|//*/get|//*/get-config')
            if root is None or len(root) == 0:
                return
            xml_type = root[0].tag
            if xml_type != "config":
                root = rpc_root.xpath('//*/filter')
            if root is None or len(root) == 0:
                return
            root_string = etree.tostring(root[0])
            xml_dic = xmltodict.parse(root_string)
            for value in xml_dic.values():
                temp_ordered_dict = value
            xml_structure_dict[xml_type] = temp_ordered_dict
        except etree.XMLSyntaxError as syntax_exception:
            xml_structure_except(syntax_exception, file_name)
        except Exception as exception_str:
            error_write(str(exception_str))
    return xml_structure_dict


def get_operations_attr(file_name):
    """Query all nodes which have operation attribute.
    Args:
        file_name: The path of xml file.
    Returns:
        operation_dict: A dict,keys record the node's xpath,
                        value record operation attribute's value.
        for example:
            operation_dict: {'/config/interfaces/interface/ipv4/address': 'create',
                             '/config/interfaces/interface': 'merge'}
    """
    operation_dict = OrderedDict()
    xml_str_without_xmlns = remove_namespace(file_name)
    if xml_str_without_xmlns != '':
        try:
            rpc_root = etree.fromstring(xml_str_without_xmlns.encode('utf-8')).getroottree()
            operation_elements = rpc_root.xpath('//*[@operation]')
            if operation_elements:
                for operation_element in operation_elements:
                    operation_xpath = rpc_root.getelementpath(operation_element)
                    operation_xpath = operation_xpath[len("edit-config"):]
                    operation_dict[operation_xpath] = operation_element.get('operation')
        except Exception as error_str:
            error_write(str(error_str))

    return operation_dict


def get_business_ordered_dict(file_path):
    result = OrderedDict()
    try:
        with open(file_path, 'r') as file_read:
            xml_ordered_dict = xmltodict.parse(file_read.read(), process_namespaces=True, xml_attribs=False)
        rpc_node = "{0}:{1}".format(COMMON_XMLNS_MAP["nc"], "rpc")
        edit_config_node = "{0}:{1}".format(COMMON_XMLNS_MAP["nc"], "edit-config")
        get_node = "{0}:{1}".format(COMMON_XMLNS_MAP["nc"], "get")
        get_config_node = "{0}:{1}".format(COMMON_XMLNS_MAP["nc"], "get-config")
        config_node = "{0}:{1}".format(COMMON_XMLNS_MAP["nc"], "config")
        filter_node = "{0}:{1}".format(COMMON_XMLNS_MAP["nc"], "filter")
        if xml_ordered_dict.get(rpc_node):
            if xml_ordered_dict[rpc_node].get(edit_config_node) and \
                    xml_ordered_dict[rpc_node][edit_config_node].get(config_node):
                result = OrderedDict([(config_node, xml_ordered_dict[rpc_node][edit_config_node][config_node])])
            elif xml_ordered_dict[rpc_node].get(get_node) and \
                    xml_ordered_dict[rpc_node][get_node].get(filter_node):
                result = OrderedDict([(filter_node, xml_ordered_dict[rpc_node][get_node][filter_node])])
            elif xml_ordered_dict[rpc_node].get(get_config_node) and \
                    xml_ordered_dict[rpc_node][get_config_node].get(filter_node):
                result = OrderedDict([(filter_node, xml_ordered_dict[rpc_node][get_config_node][filter_node])])
    except ExpatError as except_obj:
        xml_structure_except(except_obj, file_path)
    except Exception as error_mes:
        error_write(error_mes)
    finally:
        return result

def is_sub_xml(instance_xml_file, full_xml_file):
    """Determines whether instance_xml_file is a subElement of full_xml_file.
    Args:
        instance_xml_file: The path of instance-xml file.
        full_xml_file: The path of full-xml file.
    Returns:
        Bool.The result of comparison.
    """
    def _compare_ordered_dict(xml_ordered_dict, full_xml_ordered_dict):
        """In-depth traversal first principle, traversing instance_xml.(recursive function).
        Args:
            xml_ordered_dict: The OrderedDict of instance_xml.
            full_xml_ordered_dict: The OrderedDict of full_xml.
        Returns:
            result: Bool.The result of comparison.
        """
        result = True
        if result and isinstance(xml_ordered_dict, OrderedDict):
            for key, value in xml_ordered_dict.items():
                if "@" in key or "#" in key:
                    continue
                elif key in full_xml_ordered_dict.keys():
                    result = _compare_ordered_dict(xml_ordered_dict[key], full_xml_ordered_dict[key])
                else:
                    result = False
                    break
        return result

    compare_result = False
    try:
        instance_xml_order_dict = get_business_ordered_dict(instance_xml_file)
        full_xml_order_dict = get_business_ordered_dict(full_xml_file)
        compare_result = _compare_ordered_dict(instance_xml_order_dict, full_xml_order_dict)
    except Exception as except_str:
        error_write(str(except_str))

    return compare_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import os
import sys
import math
import logging
import codecs
import re

from . import pyang_util
from .. import base_util



class YangParser():
    """
    tool entrypoint class, for parse yang files use pyang in give yang_dir
    """
    parsed_datas = {}

    def __init__(self, yang_dir, features):
        self.yang_dir = yang_dir
        self.features = features
        self.ctx = pyang_util.init_ctx(yang_dir)

    # get yang files need to be parsed by self.features
    def get_yang_files(self):
        """Get the yang files that need to be parsed.
        Args:
            self.yang_dir: The path of yang file.
            self.features: A set(). Saving the namespace of module that need to be parsed.
        Returns:
            selected_files: The yang files that need to be parsed.
            Traversing yang files under the yang path,if the namespace in self.features,saving this yang file's name to
            selected_files.
        """
        selected_files = dict()
        for yang_file in os.listdir(self.yang_dir):
            try:
                if not os.path.isfile(os.path.join(self.yang_dir, yang_file)) or not yang_file.endswith(".yang"):
                    continue
                if sys.version < '3':
                    handle = codecs.open(os.path.join(self.yang_dir, yang_file), encoding="utf-8")
                else:
                    handle = open(os.path.join(self.yang_dir, yang_file), encoding="UTF-8")
                content = handle.read()
                # remove the remarks from the yang file.(The single line comment follow with code that can not remove)
                remove_comment_content = re.sub(r'\s+//[\s\S]*?\n|/\*{1,2}[\s\S]*?\*/|\s+description\s*"[^"]*?";',
                                                "", content)
                # get the yang module which need to be parsed.
                searchObj = re.search( r'\s+namespace\s+(.*?);', remove_comment_content, re.S)
                if searchObj:
                    xmlns = ""
                    # special scence: namespace "urn:ietf:params:xml:ns:"\n + "yang:ietf-isis-srv6";
                    split_result = searchObj.group(1).strip('"').split('"')
                    for item in split_result[::2]:
                        xmlns += item
                    if xmlns in self.features:
                        selected_files[os.path.splitext(yang_file)[0]] = xmlns

                handle.close()

            except (IOError, UnicodeDecodeError) as e:
                logging.error("parse yang file %s failed: %s", yang_file, e)

        return selected_files

    def parse(self):
        yang_files = self.get_yang_files()
        parser = self.parse_yang_files(yang_files)
        parser.ctx.validate()
        return parser

    def parse_yang_files(self, files):
        """
        We create only one instance to store parsed result. And we avoid of duplicate parse by record the
        yang file we have parsed and skip it in after parse.
        """
        average_progress = 0
        if files:
            completed_progress = 0
            average_progress = int(math.floor((49 - completed_progress) / len(files)))
            if average_progress == 0:
                average_progress = 1
        for yang_file, _ in files.items():
            yang_file += ".yang"
            logging.info("parse yang file %s", yang_file)
            path = os.path.join(self.yang_dir, yang_file)
            pyang_util.parse_yang_module(path, self.ctx)
            # progress_bar
            completed_progress += average_progress
            if completed_progress > 49:
                completed_progress = 49
            process_message = " {0} parse Completed.".format(yang_file)
            base_util.print_progress_bar(completed_progress, process_message)

        # modules = {}
        # main_yangs = [os.path.splitext(base_name)[0]
        #               for base_name in files.keys()]
        #
        # if self.ctx.modules:
        #     for t, stmt in self.ctx.modules.items():
        #         if t[0] in main_yangs:
        #             modules[(files[t[0]], t[0], t[1])] = stmt
        #         else:
        #             modules[(None, t[0], t[1])] = stmt
        #
        # self.ctx.modules = modules

        return self



__author__ = 'ansible_team'

from collections import OrderedDict
import os
import time
import logging
import re
from .utils.xml_parse.xml_parser import xml_to_ordered_dict
from .utils.yang_parse.yang_parser import YangParser
from .utils.yang_parse.interfaces import get_leaf_info_for_doc, get_leafinfos_from_xml_dict, \
    get_module_description, make_argument_spec, get_key_leafs, get_all_lists, check_all_node_exists
from .utils.base_util import error_write

YANG_HANDLER = ''  # define the global YANG_HANDLER
DEFAULT_INDENT = ' ' * 4


def get_xml_dict(xml_file):
    """
    get the xml_params of the xml
    :param xml_file: xml
    :return: the xml_params of the xml
    """
    return xml_to_ordered_dict(xml_file)


# get all xmlns from xxx_full.xml
def get_features_namespace(xml_dir):
    """
    :param xml_dir: the xml dir.
    :return:features_namespace: the set of all xmlns which need to be parsed.
    """
    features_namespace = set()
    files = os.listdir(xml_dir)
    for file in files:
        if os.path.isfile(os.path.join(xml_dir, file)):
            if not file.endswith("_example.xml"):
                with open(os.path.join(xml_dir, file), 'r') as xml_file:
                    xml_str = xml_file.read()
                if xml_str.strip() == '':
                    logging.error("%s is empty", file)
                else:
                    xmlns = list()
                    for item in re.findall(r'xmlns.*?="(.*?)"', xml_str):
                        xmlns.append(item)
                    for item in re.findall(r"xmlns.*?='(.*?)'", xml_str):
                        xmlns.append(item)
                    if xmlns:
                        for xmlns_item in xmlns:
                            features_namespace.add(xmlns_item)
                    else:
                        logging.error("%s has no xmlns", file)
        else:
            features_namespace = features_namespace | get_features_namespace(os.path.join(xml_dir, file))

    return features_namespace


def get_yang_handlers(yang_dir, xml_dir):
    """
    get the yang_handler of the yang_dir
    :param dir: the dir containing yang files
    :return:YANG_HANDLER
    """
    global YANG_HANDLER

    features = get_features_namespace(xml_dir)
    factory = YangParser(yang_dir, features)
    YANG_HANDLER = factory.parse()
    return YANG_HANDLER


def get_xml_descption(xml_namespace):
    """
    get the xml_description from yang_file of the xml_file
    :param xml_file: xml_file
    :return: the xml_description
    """
    parser = YANG_HANDLER
    return get_module_description(parser, xml_namespace)


def get_xml_options(full_xml_file_path, full_xml_ordered_dict, xmlns_info):
    """
    get the xml_params of the xml
    :param xml_file: xml_file
    :return:the xml_params of the xml
    """
    return get_leaf_info_for_doc(full_xml_file_path, full_xml_ordered_dict, YANG_HANDLER, xmlns_info)


def get_xml_leafinfos(full_xml_ordered_dict, xmlns_info):
    """
    get the xml_leafinfos of the xml
    :param xml_file: xml_file
    :return:the xml_leafinfos of the xml
    """
    yang_handlers = YANG_HANDLER
    xml_leafinfos = get_leafinfos_from_xml_dict(full_xml_ordered_dict, yang_handlers, xmlns_info)
    return xml_leafinfos


def get_xml_key_leafs(full_xml_ordered_dict, xmlns_info):
    """
    get the xml_leafinfos of the xml
    :param full_xml_ordered_dict:
    :return:the xml_leafinfos of the xml
    """
    yang_handlers = YANG_HANDLER
    xml_key_leafs = get_key_leafs(full_xml_ordered_dict, yang_handlers, xmlns_info)
    return xml_key_leafs


def get_xml_list_leafs(full_xml_ordered_dict, xmlns_info):
    """
    get the xml_leafinfos of the xml
    :param full_xml_ordered_dict:
    :return:the xml_leafinfos of the xml
    """
    yang_handlers = YANG_HANDLER
    xml_key_leafs = get_all_lists(full_xml_ordered_dict, yang_handlers, xmlns_info)
    return xml_key_leafs


def gen_argument_spec(full_xml_ordered_dict, xmlns_info, key_list, list_list):
    """
    get the argument_spec of the xml
    :param xml_leafinfos: xml_leafinfos
    :return:the argument_spec of the xml
    """
    return make_argument_spec(full_xml_ordered_dict, YANG_HANDLER, xmlns_info, key_list, list_list)


def gen_options(data, options_str, level=0):
    level += 1
    if isinstance(data, OrderedDict):
        for key, value in data.items():
            options_str += DEFAULT_INDENT + DEFAULT_INDENT * (level - 1) + key + ':' + '\n'
            if isinstance(value, tuple):
                if not value:
                    error_write("node %s has no value" % key)
                    return ""
                options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'description:\n'
                options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + DEFAULT_INDENT + '- ' + \
                               word_wrap(str(value[0]), 4 * (level + 2) + 3) + '\n'
                for when_must_check in value:
                    if when_must_check == "when:True":
                        str_out = "when: The configuration of this object takes effect only when certain conditions are met. For details, check the definition in the YANG model."
                        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + str_out + '\n'
                    elif when_must_check == "must:True":
                        str_out = "must: The configuration of this object takes effect only when certain conditions are met. For details, check the definition in the YANG model."
                        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + str_out + '\n'
                options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'required:' + str(value[1]) + '\n'
                if isinstance(value[-1], OrderedDict):
                    options_str += gen_options(value[-1], '', level)
                else:
                    options_str = option_check(options_str, value, level)
            elif isinstance(value, OrderedDict):
                options_str += gen_options(value, '', level)
    return options_str


def option_check(options_str, value, level):
    """
        get the documention of the value
        :return:the documention
        """
    if value[7] == True:
        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'key:' + str(value[7]) + '\n'
    if value[-4] == True:
        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'mandatory:' + str(value[8]) + '\n'
    if value[-2] == True:
        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'suppor-filter:' + str(value[10]) + '\n'
    if value[2] != None:
        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'default:' + str(value[2]) + '\n'
    if value[3] != []:
        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'pattern:' + str(value[3]) + '\n'
    options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'type:' + str(value[-1]) + '\n'
    if value[-1] == 'str' and value[6] != []:
        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'length:' + str(value[6]) + '\n'
    if value[-1] == 'int':
        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'range:' + str(value[5]) + '\n'
    if value[-1] in ['enum', 'bool']:
        options_str += DEFAULT_INDENT + DEFAULT_INDENT * level + 'choices:' + str(value[4]) + '\n'
    return options_str


def line_wrap(desc, indent):
    wrapped_desc = []
    length = len(desc)
    line_len = 160 - indent
    while length > line_len:
        line = desc[0:line_len]
        if not line[-1].isspace():
            idx = line_len - 1
            while not line[idx].isspace():
                idx = idx - 1
            line = desc[:idx + 1]
            desc = desc[idx + 1:]
        else:
            desc = desc[line_len:]
        wrapped_desc.append(line + "\n")
        length = len(desc)
        if length <= line_len:
            wrapped_desc.append(desc)
    if not wrapped_desc:
        wrapped_desc.append(desc)

    return (" " * (indent - 1)).join(wrapped_desc)


def word_wrap(description, indent):
    descs = [section for section in description.split("\n") if section.strip()]
    wrapped_descs = []
    first = True
    for desc in descs:
        wrapped_desc = line_wrap(desc, indent)
        if not first:
            wrapped_desc = (" " * (indent - 1)) + wrapped_desc
        first = False
        wrapped_descs.append(wrapped_desc)

    return "\n".join(wrapped_descs)


def gen_documentation(module_desc, data, filename, xml_type):
    '''
    :param module_desc: module description
    :param data: xml_yang data
    :return:documentation
    '''
    documentation = ''
    documentation += '---' + '\n'
    documentation += 'module:' + filename + '\n'
    documentation += 'version_added: "2.6"' + '\n'
    documentation += 'short_description: ' + word_wrap(module_desc, 20) + '\n'
    documentation += 'description:' + '\n'
    documentation += '    - ' + word_wrap(module_desc, 7) + '\n'
    documentation += 'author:' + (__author__) + '@huawei' + '\n'
    now_time = (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(int(time.time()))))
    documentation += 'time:' + str(now_time) + '\n'
    documentation += 'options:\n'
    options_str = get_options_str(xml_type)
    documentation += gen_options(data, options_str)
    try:
        return documentation
    except Exception as e:
        print(e)


def get_options_str(xml_type):
    options_str = ''
    options_str += DEFAULT_INDENT + DEFAULT_INDENT * 0 + 'opreation_type:{}\n'.format(xml_type)
    options_str += DEFAULT_INDENT + DEFAULT_INDENT * 1 + 'description:{}\n'.format(xml_type)
    options_str += DEFAULT_INDENT + DEFAULT_INDENT * 1 + DEFAULT_INDENT + '- ' + \
                   word_wrap('This is a helper node ,Choose from config, get', 4 * 3 + 3) + '\n'
    options_str += DEFAULT_INDENT + DEFAULT_INDENT * 1 + 'type: str\n'
    options_str += DEFAULT_INDENT + DEFAULT_INDENT * 1 + 'required:{}\n'.format(True)
    options_str += DEFAULT_INDENT + DEFAULT_INDENT * 1 + 'choices: ["config","get","get-config","input_action"]\n'
    return options_str


def check_xpaths_exists(xml_dict, xmlns_info):
    return check_all_node_exists(xml_dict, YANG_HANDLER, xmlns_info)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""pyang_wrapper.py docstrings.
This module invoke mpyang.
"""
import re
import logging
import os
import io

from pyang.repository import FileRepository
from pyang.context import Context

class MockOpts:
    """
    MockOpts class.
    """
    def __init__(self):
        self.format = 'yang'
        self.yang_remove_unused_imports = False
        self.yang_canonical = True
        self.yin_canonical = True
        self.trim_yin = False
        self.yin_pretty_strings = True
        self.outfile = None
        self.xx = 0


def init_ctx(yang_dir):
    """
    init_ctx.
    """
    path = yang_dir
    repos = FileRepository(path)
    ctx = Context(repos)
    opts = MockOpts()
    ctx.opts = opts
    ctx.trim_yin = opts.trim_yin
    return ctx


class YangContext:
    """
    YangContext class.
    """
    def __init__(self, yang_dir):
        self.ctx = None
        self.files = set()
        try:
            self.ctx = init_ctx(yang_dir)
        except Exception:
            logging.error("get_module_context_pieces error")

    def parse_yang_module(self, filename):
        """
        parse_yang_module.
        """
        ctx = self.ctx
        in_base_name = os.path.basename(filename)
        if in_base_name in self.files:
            logging.info(
                "%s has been parsed, no need parse again", in_base_name)
        file_open = None
        yang_file_regular = re.compile(r"^(.*?)(\@(\d{4}-\d{2}-\d{2}))?\.(yang|yin)$")
        try:
            file_open = io.open(filename, "r", encoding="utf-8")
            text = file_open.read()
            yang_file = yang_file_regular.search(filename)
            ctx.yin_module_map = {}
            if yang_file is not None:
                (name, _dummy, rev, format) = yang_file.groups()
                name = os.path.basename(name)
                ctx.add_module(filename, text, format, name, rev,
                               expect_failure_error=False)
            else:
                ctx.add_module(filename, text)
        except Exception:
            logging.error("can not open the file %s" % filename)
        finally:
            if file_open is not None:
                file_open.close()
            if ctx.modules:
                self.files = self.files | set(
                    [t[0] for t in ctx.modules.keys()])



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import copy
import logging
from collections import OrderedDict

from . import pyang_util
from . import constant
from . import utils
from ..xml_parse.xml_parser_get_xmlns import get_node_xmlns
from ..base_util import operation_warning_write
from .constant import BASE_INTEGER_TYPES

LIST_KEY_SET = set()

def get_feature_info(xpath, parsed_data, xmlns_info):
    """Get the module from yang_handler.
    Args:
        xpath: The xpath-node which need to find module.(without prefix).
        parsed_data: The yang_handler.
        xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
    Returns:
        feature_info: The module that xpath-node belong to.
        augment_flag: Whether the xpath-node is augment.
    """
    augment_flag = False
    # get xpath-node's xmlns.
    current_node_xmlns = get_node_xmlns(xpath, xmlns_info)

    feature_xpath = "/" + xpath.split("/")[1]
    xml_namespace = get_node_xmlns(feature_xpath, xmlns_info)

    # whether augment: xpath-node's xmlns not equal xml's xmlns.
    if current_node_xmlns != xml_namespace:
        augment_flag = True
    # use xpath-node's xmlns to get the feature_info.
    feature_info = None
    augment_break_out_flag = False
    for _, module in parsed_data.ctx.modules.items():
        if module.keyword == 'module':
            if augment_flag:
                if augment_break_out_flag:
                    break
                else:
                    val_namespace2 = module.search_one("namespace")
                    if val_namespace2 and val_namespace2.arg == current_node_xmlns:
                        for _augment, module_by_augment in module.i_ctx.modules.items():
                            val_namespace2 = module_by_augment.search_one("namespace")
                            if val_namespace2 and val_namespace2.arg == xml_namespace:
                                feature_info = module_by_augment
                                augment_break_out_flag = True
                                break
            else:
                val_namespace = module.search_one("namespace")
                if val_namespace and val_namespace.arg == xml_namespace:
                    feature_info = module
                    break
    return [feature_info, augment_flag]

def recursive_root(node_name, root):
    """Recursive to get xpath-node's root.(recursive function).
    Args:
        node_name: The node name that split by xpath.
        root: The module's root.
    Returns:
        root: The xpath-node's root.
    """
    if not hasattr(root, 'i_children'):
        return None
    for node in root.i_children:
        if node.keyword in  ['input','output','choice','case']:
            sub_root = recursive_root(node_name,node)
            if(sub_root is not None):
                return sub_root
        else:
            if (node.arg == node_name):
                return node
    return None

def get_xpaths_node(xpath, root):
    """Get the xpath-node from feature_info.
    Args:
        xpath: The xpath-node which need to find root.(without prefix).
        root: The module's root that xpath-node belong to.
    Returns:
        root: The xpath-node's root.
    """
    for node_name in xpath.split("/")[1:]:
        if root is None:
            break
        else:
            root = recursive_root(node_name, root)

    if root is None:
        return None
    else:
        return root

def recursive_root_of_augment(node_name, root):
    """Recursive to get xpath-node's root when augment.(recursive function).
    Args:
        node_name: The node name that split by xpath.
        root: The module's root.
    Returns:
        root: The xpath-node's root.
    """
    if not hasattr(root, 'i_children'):
        return None
    current_node_stmts = [node for node in root.i_children]
    for current_node in current_node_stmts:
        if current_node.keyword in ('input','output','choice','case'):
            root = recursive_root_of_augment(node_name,current_node)
            if root is not None:
                return root
            else:
                continue
        if node_name == current_node.arg:
            root = current_node
            return root
    return root

def get_xpaths_node_of_augment(xpath, root):
    """Get the xpath-node from feature_info when augment.
    Args:
        xpath: The xpath-node which need to find root.(without prefix).
        root: The module's root that xpath-node belong to.
    Returns:
        root: The xpath-node's root.
    """
    for node_name in xpath.split("/")[1:]:
        if root is None:
            break
        else:
            root = recursive_root_of_augment(node_name, root)

    if root is None:
        return None
    else:
        return root


def get_required(node):
    """Get xpath-node required information.
    Args:
        node: The xpath-node's root.
    Returns:
        A bool value.
        rule: The xpath-node that has a "mandatory" attr or is primary-key,it is required.else not required.
    """
    keys = []
    p_stmt = node.parent
    key_stmt = p_stmt.search_one("key")
    if key_stmt:
        keys = [name.strip() for name in key_stmt.arg.split() if name.strip()]
    if node.arg in keys:
        return True
    mandatory_stmt = node.search_one("mandatory")
    if mandatory_stmt and mandatory_stmt.arg.strip().lower() == "true":
        return True
    # when node is a case,the mandatory exist in parent's substmts.
    if node.arg == node.parent.arg:
        if node.parent.keyword == "case":
            if node.parent.parent.keyword == "choice":
                for choice_substmts in node.parent.parent.substmts:
                    if choice_substmts.keyword == "mandatory" and choice_substmts.arg.strip().lower() == "true":
                        return True
    return False

def get_default(node):
    """Get xpath-node default information.
    Args:
        node: The xpath-node's root.
    Returns:
        return xpath-node's default value.
    """

    default_stmt = node.search_one("default")
    if default_stmt:
        default = default_stmt.arg
        type_stmt = node.search_one("type")
        if not type_stmt.i_type_spec:
            return default
        if type_stmt.i_type_spec.name == "boolean":
            if default.lower() == "false":
                default = False
            elif default.lower() == "true":
                default = True
        elif type_stmt.i_type_spec.name in constant.BASE_INTEGER_TYPES:
            default = int(default)
        return default
    else:
        return None

def get_type(node):
    """Get xpath-node's type.(recursive function).
    Args:
        node: The xpath-node's root.
    Returns:
        return xpath-node's type.
    """
    if node.keyword not in ['leaf','leaf-list']:
        return None
    type_stmt = node.search_one("type")
    if not type_stmt:
        logging.info("leaf %s has no type defination", node.arg)
        return None
    type_spec = type_stmt.i_type_spec
    if not type_spec:
        return None
    if type_spec.name == "leafref":
        if hasattr(type_spec, "i_target_node"):
            target_node = type_spec.i_target_node
            return get_type(target_node)
        else:
            return None
    else:
        return type_stmt.i_type_spec.name

def get_pattern(node):
    """Get xpath-node's pattern.(recursive function).
    Args:
        node: The xpath-node's root.
    Returns:
        return xpath-node's pattern.
    """
    if node.keyword not in ['leaf','leaf-list']:
        return None
    type_stmt = node.search_one("type")
    if not type_stmt:
        logging.info("leaf %s has no type defination", node.arg)
        return None
    type_spec = type_stmt.i_type_spec
    if not type_spec:
        return []
    patterns = []
    if hasattr(type_stmt.i_type_spec, "res"):
        patterns.extend([pattern_stmt.spec for pattern_stmt in type_stmt.i_type_spec.res])
    elif type_spec.name == "leafref":
        patterns.extend(get_pattern(type_spec.i_target_node))
    return patterns

def get_restrict(node):
    """Get xpath-node's restrict.(recursive function).
    Args:
        node: The xpath-node's root.
    Returns:
        return xpath-node's restrict.
    """
    if node.keyword not in {'leaf','leaf-list'}:
        return None
    type_stmt = node.search_one("type")
    if not type_stmt:
        logging.info("leaf %s has no type defination", node.arg)
        return None
    type_spec = type_stmt.i_type_spec
    if not type_spec:
        return []
    if type_spec.name == "leafref":
        if not hasattr(type_spec,'i_target_node'):
            logging.error("leaf %s type is %s but no target node", node.arg,type_spec.name)
            return None
        target_node = type_spec.i_target_node
        return get_restrict(target_node)
    if type_spec.name == "string":
        rests = []
        if hasattr(type_spec, "lengths"):
            rests.extend(type_stmt.i_type_spec.lengths)
        base_def = type_spec.base
        while base_def:
            if hasattr(base_def, "lengths"):
                rests.extend(base_def.lengths)
            base_def = base_def.base
        return utils.Util.restrict_set_intersec(rests)
    elif type_spec.name == "enumeration":
        return [enum_t[0] for enum_t in type_spec.enums]
    elif type_spec.name in constant.BASE_INTEGER_TYPES:
        rests = []
        if hasattr(type_spec, "min") and hasattr(type_spec, "max"):
            rests.extend([(type_spec.min, type_spec.max)])
        elif hasattr(type_spec, "ranges"):
            rests.extend(type_spec.ranges)
        base_def = type_spec.base
        while base_def:
            if hasattr(type_spec, "min") and hasattr(type_spec, "max"):
                rests.extend([(type_spec.min, type_spec.max)])
            elif hasattr(base_def, "ranges"):
                rests.extend(base_def.ranges)
            base_def = base_def.base
        return utils.Util.restrict_set_intersec(rests)
    else:
        return []

def get_desc(node):
    """Get xpath-node's description.
    Args:
        node: The xpath-node's root.
    Returns:
        return xpath-node's description.
    """
    desc_stmt = node.search_one("description")
    if desc_stmt:
        return desc_stmt.arg
    else:
        return ""

def check_is_key(node):
    """Check the xpath-node whether is primary-key.
    Args:
        node: The xpath-node's root.
    Returns:
        A bool.
    """
    parent_stmt = node.parent
    key_stmt = parent_stmt.search_one("key")
    if key_stmt:
        a = node.arg in [name.strip() for name in key_stmt.arg.split() if name.strip()]
        return node.arg in [name.strip() for name in key_stmt.arg.split() if name.strip()]
    return False


def check_config(node):
    result = True
    config_stmt = node.search_one("config")
    if config_stmt and config_stmt.arg.lower() == "false":
        result = False
    return result
def when_must_check(node):
    """Get xpath-node default information.
       Args:
           node: The xpath-node's root.
       Returns:
           return xpath-node's default value.
       """
    when_stmt = node.search_one("when")
    must_stmt = node.search_one("must")
    if when_stmt:
        return "when:True"
    elif must_stmt:
        return "must:True"
    else:
        return None
def mandatory_check(node):
    """Get xpath-node default information.
       Args:
           node: The xpath-node's root.
       Returns:
           return xpath-node's default value.
       """
    when_stmt = node.search_one("mandatory")
    if when_stmt:
        return True
    else:
        return None

def suport_filter_check(node):
    """Get xpath-node default information.
       Args:
           node: The xpath-node's root.
       Returns:
           return xpath-node's default value.
       """
    support_filter = ('huawei-extension', 'support-filter')
    when_stmt = node.search_one(support_filter)
    if when_stmt:
        return True
    else:
        return None

def get_xpaths_infos(xpath, parsed_data, xmlns_info):
    """Get xpath-node information from yang_handler.
    Args:
        xpath: The xpath-node which need to get information.(without prefix).
        parsed_data: The yang_handler.
        xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
    Returns:
        A dict.Saving xpath-node's information.
        for example:
            {'key': False,
             'desc': 'Network instance to which an interface is bound.',
             'restrict': [],
             'pattern': [],
             'required': False,
             'default': None,
             'type': 'string'
             }
    """

    # get_xpaths_infos() begin.
    if not parsed_data.ctx.modules:
        return {}
    # get module's root and augment_flag.
    feature_info_list = get_feature_info(xpath, parsed_data, xmlns_info)
    feature_info = feature_info_list[0]
    augment_flag = feature_info_list[1]

    feature = xpath.split("/")[1]
    if not feature_info:
        logging.error("module %s is not provided in yang directory", feature)
        return {}

    if augment_flag:
        # get xpath-node's root when augment.
        leaf = get_xpaths_node_of_augment(xpath, feature_info)
    else:
        # get xpath-node's root.
        leaf = get_xpaths_node(xpath, feature_info)

    if leaf:
        return {
            'required': get_required(leaf),
            'type': get_type(leaf),
            'restrict': get_restrict(leaf),
            "pattern": get_pattern(leaf),
            'default': get_default(leaf),
            'desc': get_desc(leaf),
            'key': check_is_key(leaf),
            'whether_config': check_config(leaf),
            'when_must_check':when_must_check(leaf),
            'mandatory_check':mandatory_check(leaf),
            'suport_filter_check':suport_filter_check(leaf)
        }
    else:
        return {}

def get_module_description(parser, xml_namespace):
    """Get module's description from yang_handler.
    Args:
        parser: The yang_handler.
        xml_namespace: The xml's namespace.
    Returns:
        first we try to get description statement in main module yang files, if there is no description
        statement, then we try to get description statement from root node.
    """
    result = ""
    feature_info_list = []
    if len(xml_namespace) == 0:
        return result
    # use the xml_namespace to find module.
    for key, val in parser.ctx.modules.items():
        if not val.keyword == 'module':
            continue
        module_ns = pyang_util.get_module_namespace(val)
        if module_ns in xml_namespace:
            feature_info_list.append(val)
    if not feature_info_list:
        logging.error("module %s is not provided in yang directory", xml_namespace)
        return result

    feature_info_list.sort(key=lambda x: x.arg)
    for feature_info in feature_info_list:
        desc_stmt = feature_info.search_one("description")
        if desc_stmt and desc_stmt.arg:
            result = result + desc_stmt.arg + "\n"
    return result


def check_all_node_exists(xml_dict, yang_parser, xmlns_info):
    """Check all node exists in yang_handler.
    Args:
        xml_dict: An OrderedDict. Recording the xml's structure.
        yang_parser: The yang_handler.
        xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
    Returns:
        A bool.
    """

    def gen_xpaths(xpaths,xml_dict):
        if xml_dict is None:
            return  xpaths;
        if xpaths is None or len(xpaths) ==0 :
            last_xpath = ""
        else:
            last_xpath = xpaths[-1]
        for key,val in xml_dict.items():
            xpath = last_xpath + "/" +  key
            xpaths.append(xpath)
            if not isinstance(val, list):
                nodes = [val]
            else:
                nodes = val
            for node in nodes:
                gen_xpaths(xpaths,node)

    # check_all_node_exists() begin.
    xpaths=[]
    gen_xpaths(xpaths,list(xml_dict.values())[0])

    for xpath in xpaths:
        xpath_info = get_xpaths_infos(xpath,yang_parser,xmlns_info)
        if not xpath_info:
            logging.warning("xpath %s not exist in yang files", xpath)
            return False
    return True

def is_key_leaf(xpath, yang_parser, xmlns_info):
    """Whether the leaf-node is primary-key
    Args:
        xpath: The leaf-node's xpath.(without prefix).
        yang_parser: The yang_handler.
        xmlns_info:  The xmlns_info.Saving all nodes's xmlns information.
    Returns:
        A bool.
    """
    global LIST_KEY_SET
    if xpath in LIST_KEY_SET:
        return True

    feature = xpath.split("/")[1]
    if not yang_parser.ctx.modules:
        return False

    # get module's root and augment_flag.
    feature_info_list = get_feature_info(xpath, yang_parser, xmlns_info)
    feature_info = feature_info_list[0]
    augment_flag = feature_info_list[1]

    if not feature_info:
        logging.error("module %s is not provided in yang directory", feature)
        return False

    if augment_flag:
        # get xpath's root when augment.
        leaf = get_xpaths_node_of_augment(xpath, feature_info)
    else:
        # get xpath's root.
        leaf = get_xpaths_node(xpath, feature_info)

    if not leaf:
        logging.warning("xpath %s has no correspond leaf find in yang files", xpath)
        return False

    if leaf.keyword == "list":
        key_stmt = leaf.search_one("key")
        if key_stmt:
            for name in key_stmt.arg.split():
                if name.strip():
                    LIST_KEY_SET.add(str(xpath) + "/" + str(name))

    return False

def is_list_node(xpath, yang_parser, xmlns_info):
    """Whether the node is list
    Args:
        xpath: The leaf-node's xpath.(without prefix).
        yang_parser: The yang_handler.
        xmlns_info:  The xmlns_info.Saving all nodes's xmlns information.
    Returns:
        A bool.
    """
    feature = xpath.split("/")[1]
    if not yang_parser.ctx.modules:
        return False

    # get module's root and augment_flag.
    feature_info_list = get_feature_info(xpath, yang_parser, xmlns_info)
    feature_info = feature_info_list[0]
    augment_flag = feature_info_list[1]

    if not feature_info:
        logging.error("module %s is not provided in yang directory", feature)
        return False

    if augment_flag:
        # get xpath's root when augment.
        leaf = get_xpaths_node_of_augment(xpath, feature_info)
    else:
        # get xpath's root.
        leaf = get_xpaths_node(xpath, feature_info)

    if not leaf:
        logging.warning("xpath %s has no correspond leaf find in yang files", xpath)
        return False

    if leaf.keyword == "list":
        return True

    return False


def get_key_leafs(xml_dict, yang_parser, xmlns_info):
    """Traversing the xml_dict,get all leaf-node's xpath which is primary-key.
    Args:
        xml_dict: An OrderedDict. Recording the xml's structure.
        yang_parser: The yang_handler.
        xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
    Returns:
        A list.
    """
    def get_keys(leaf_dict, xpath, yang_parser, leaf_path, xmlns_info):
        """Get all leaf-node's xpath which is primary-key.(recursive function).
        Args:
            leaf_dict: An OrderedDict. Recording the container-node's structure.
            xpath: container-node's xpath.
            yang_parser: The yang_handler.
            leaf_path: Used to record the leaf-node's xpath.
            xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
        Returns:
            A list.
        """
        if leaf_dict is None:
            return None
        keys = []
        for key, val in leaf_dict.items():
            container_leaf_xpath = leaf_path + "/" + key
            tmp_xpath = xpath + "/" + key
            if is_key_leaf(tmp_xpath, yang_parser, xmlns_info):
                #logging.info("xpath %s is key leaf", tmp_xpath)
                keys.append(leaf_path + "/" + key)
            if isinstance(val, OrderedDict):
                keys.extend(
                    get_keys(leaf_dict[key], tmp_xpath, yang_parser, container_leaf_xpath, xmlns_info))
        return keys

    # get_key_leafs() begin.
    tmp_dict = copy.deepcopy(list(xml_dict.values())[0])
    xpath = ""
    leaf_path = ""
    global LIST_KEY_SET
    LIST_KEY_SET = set()
    key_leafs = get_keys(tmp_dict, xpath, yang_parser, leaf_path, xmlns_info)

    return key_leafs

def get_all_lists(xml_dict, yang_parser, xmlns_info):
    """Traversing the xml_dict,get all list's xpath.
    Args:
        xml_dict: An OrderedDict. Recording the xml's structure.
        yang_parser: The yang_handler.
        xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
    Returns:
        A list.
    """
    def get_lists(leaf_dict, xpath, yang_parser, leaf_path, xmlns_info):
        """Get all list's xpath which is primary-key.(recursive function).
        Args:
            leaf_dict: An OrderedDict. Recording the container-node's structure.
            xpath: container-node's xpath.
            yang_parser: The yang_handler.
            leaf_path: Used to record the leaf-node's xpath.
            xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
        Returns:
            A list.
        """
        all_lists = []
        if leaf_dict is None:
            return all_lists
        for key, val in leaf_dict.items():
            tmp_xpath = xpath
            container_leaf_xpath = leaf_path + "/" + key
            if isinstance(val, OrderedDict):
                tmp_xpath = tmp_xpath + "/" + key
                if is_list_node(tmp_xpath, yang_parser, xmlns_info):
                    all_lists.append(leaf_path + "/" + key)
                all_lists.extend(
                    get_lists(leaf_dict[key], tmp_xpath, yang_parser, container_leaf_xpath, xmlns_info))
            else:
                tmp_xpath = tmp_xpath + "/" + key
                if is_list_node(tmp_xpath, yang_parser, xmlns_info):
                    all_lists.append(leaf_path + "/" + key)
        return all_lists

    # get_key_leafs() begin.
    tmp_dict = copy.deepcopy(list(xml_dict.values())[0])
    xpath = ""
    leaf_path = ""
    all_lists = get_lists(tmp_dict, xpath, yang_parser, leaf_path, xmlns_info)

    return all_lists

def get_leafinfos_from_xml_dict(node_d, parser, xmlns_info):
    """Traversing the xml_dict,get all leaf-node's information.
    Args:
        node_d: An OrderedDict. Recording the xml's structure.
        parser: The yang_handler.
        xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
    Returns:
        An OrderedDict.
        The information contain (key,desc,restrict,pattern,required,default,type) which come from each leaf-node.
        The output is for params check.
    """
    def get_leafinfos(xml_type, leaf_dict, xpath, parser, xmlns_info):
        """Get all leaf-node's information.(recursive function).
        Args:
            leaf_dict: An OrderedDict. Recording the container-node's structure.
            xpath: Used to record the leaf-node's xpath.
            parser: The yang_handler.
            xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
        Returns:
            An OrderedDict.
        """
        leaf_infos = OrderedDict()
        if leaf_dict is None:
            return leaf_infos;
        for key, val in leaf_dict.items():
            if not isinstance(val, list):
                # compatible the situation with list
                recurv_val = [val]
            else:
                recurv_val = val
            for item in recurv_val:
                if isinstance(item, OrderedDict):
                    tmp_xpath = xpath + "/" + key
                    if key not in leaf_infos:
                        leaf_infos[key] = OrderedDict()
                    leaf_infos[key].update(
                        get_leafinfos(xml_type, item, tmp_xpath, parser, xmlns_info))
                else:
                    tmp_xpath = xpath + "/" + key
                    yang_info = get_xpaths_infos(tmp_xpath, parser, xmlns_info)
                    # default = None
                    # yang_info['required'] = False
                    # default = {'default':None}
                    # yang_info.update(default)
                    if yang_info:
                        if xml_type == "config" and not yang_info.get("whether_config"):
                            continue
                        _d1 = trans_to_smiple_type(yang_info)
                        if not yang_info.__contains__('type'):
                            type = None
                        else:
                            type = yang_info["type"]
                        if not yang_info.__contains__('pattern'):
                            pattern = None
                        else:
                            pattern = yang_info["pattern"]
                        if not yang_info.__contains__('key'):
                            is_key = None
                        else:
                            is_key = yang_info["key"]
                        if not yang_info.__contains__('required'):
                            required = None
                        else:
                            required = yang_info["required"]
                        if not yang_info.__contains__('default'):
                            default = None
                        else:
                            default = yang_info["default"]

                        form_leaf = OrderedDict({
                            key: {'required': required,
                                  'type': type,
                                  "default": default,
                                  "pattern": pattern,
                                  'key': is_key
                                  }
                        })
                        form_leaf[key].update(_d1)
                        leaf_infos.update(form_leaf)
                    else:
                        logging.info(
                            "xpath %s not find in yang files", tmp_xpath)
        return leaf_infos

    xml_type = list(node_d.keys())[0]
    xml_dict = copy.deepcopy(list(node_d.values())[0])
    xpath = ""

    leaf_infos = get_leafinfos(xml_type, xml_dict, xpath, parser, xmlns_info)
    return leaf_infos


def trans_to_smiple_type(yang_info):
    if 'type' not in yang_info:
        return {}
    if yang_info['type'] in constant.BASE_INTEGER_TYPES:
        _d1 = {'range': yang_info["restrict"]}
        yang_info['type'] = 'int'
    elif yang_info['type'] == "string":
        _d1 = {'length': yang_info["restrict"]}
    elif yang_info['type'] == "enumeration":
        _d1 = {'choices': yang_info['restrict']}
    elif yang_info['type'] == "identityref":
        _d1 = {'type': "string", 'length': []}
    elif yang_info['type'] == "empty":
        _d1 = {'type': "string", 'length': []}
    elif yang_info['type'] == "union":
        _d1 = {'type': "string", 'length': []}
    else:
        _d1 = {}
    return _d1
def get_leaf_info_for_doc(full_xml_file_path, leaf_dict, parser, xmlns_info):
    """Traversing the xml_dict,get all leaf-node's information.
    Args:
        leaf_dict: An OrderedDict. Recording the xml's structure.
        parser: The yang_handler.
        xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
    Returns:
        An OrderedDict.
        The information contain (description, required , default) which come from each leaf-node.
        The output is for DOCUMENTATION's content.
    """
    def get_leafinfos(full_xml_file_path, xml_type, leaf_dict, xpath, parser, xmlns_info):
        """Get all leaf-node's information.(recursive function).
        Args:
            leaf_dict: An OrderedDict. Recording the container-node's structure.
            xpath: Used to record the leaf-node's xpath.
            parser: The yang_handler.
            xmlns_info: The xmlns_info.Saving all nodes's xmlns information.
        Returns:
            An OrderedDict.
        """
        doc_infos = OrderedDict()
        if leaf_dict is None:
            return None
        for key, val in leaf_dict.items():
            if not isinstance(val, list):
                # compatible the situation with list
                val = [val]
            if key not in doc_infos:
                doc_infos[key] = OrderedDict()
            for node in val:
                tmp_xpath = xpath
                if isinstance(node, OrderedDict):
                    tmp_xpath = tmp_xpath + "/" + key
                    if key == "@operation":
                        continue
                    doc_flag = True
                    leaf_info = get_xpaths_infos(tmp_xpath, parser, xmlns_info)
                    if leaf_info:
                        if xml_type == "config" and not leaf_info["whether_config"]:
                            msg = 'The node: %s is [config "false"],ansible-gen will ignore this node.' \
                                  '\nplease check the file: %s.' \
                                  '\nplease confirm the instance-xml synchronously.' % (tmp_xpath, full_xml_file_path)
                            operation_warning_write(msg)
                            doc_infos.pop(key)
                            continue
                        _c = put_doc_type(leaf_info)
                        doc_infos[key] = (
                            leaf_info["desc"], leaf_info["required"], leaf_info["default"],
                            leaf_info.get('pattern'),leaf_info.get('when_must_check'),
                            _c.get('type'))
                    else:
                        logging.warning(
                            "xpath %s not find in  yang files", tmp_xpath)
                        doc_infos[key] = tuple()
                    doc_infos[key] = list(doc_infos[key])
                    empty_orderdict = OrderedDict()
                    doc_infos[key].append(empty_orderdict)
                    doc_infos[key] = tuple(doc_infos[key])
                    doc_infos[key][-1].update(
                        get_leafinfos(full_xml_file_path, xml_type, node, tmp_xpath, parser, xmlns_info))
                else:
                    if key == "@operation":
                        continue
                    tmp_xpath = tmp_xpath + "/" + key
                    leaf_info = get_xpaths_infos(tmp_xpath, parser, xmlns_info)
                    if leaf_info:
                        if xml_type == "config" and not leaf_info["whether_config"]:
                            msg = 'The node: %s is [config "false"],ansible-gen will ignore this node.' \
                                  '\nplease check the file: %s.' \
                                  '\nplease confirm the instance-xml synchronously.' % (tmp_xpath, full_xml_file_path)
                            operation_warning_write(msg)
                            doc_infos.pop(key)
                            continue
                        _c = put_doc_type(leaf_info)
                        if leaf_info['type'] == 'enumeration':
                            leaf_info['choices'] = leaf_info.get('restrict')
                        if leaf_info['type'] == 'boolean':
                            leaf_info['choices'] = ['true','false']
                        if leaf_info['type'] in BASE_INTEGER_TYPES:
                            leaf_info['range'] = leaf_info.get('restrict')
                        if leaf_info['type'] == 'string':
                            leaf_info['length'] = leaf_info.get('restrict')
                        doc_infos[key] = (
                            leaf_info["desc"], leaf_info["required"], leaf_info["default"],
                            leaf_info.get('pattern'),leaf_info.get('choices'),leaf_info.get('range'),leaf_info.get('length'),leaf_info.get('key'),
                            leaf_info.get('mandatory_check'),leaf_info.get('when_must_check'),leaf_info.get('suport_filter_check'),
                            _c.get('type'))
                    else:
                        logging.warning(
                            "xpath %s not find in  yang files", tmp_xpath)
                        doc_infos[key] = tuple()
        return doc_infos

    xpath = ""
    xml_type = list(leaf_dict.keys())[0]
    xml_dict = copy.deepcopy(list(leaf_dict.values())[0])

    doc_infos = get_leafinfos(full_xml_file_path, xml_type, xml_dict, xpath, parser, xmlns_info)

    return doc_infos

def put_doc_type(leaf_info):
    """Generate the leaf-node's DOCUMENTATION.
    Args:
        leaf_infos: The result of get_leafinfos_from_xml_dict().
    Returns:
        A dict.The leaf-node's DOCUMENTATION.
        The leaf-node's DOCUMENTATION has (type, default, required) attribution which come from leaf_info.
    """
    dic = dict()
    try:
        if leaf_info['type'] in constant.BASE_INTEGER_TYPES:
            dic['type'] = 'int'
        elif leaf_info['type'] in ("union", "identityref", "empty", "string"):
            dic['type'] = 'str'
        elif leaf_info['type'] == "enumeration":
            dic['type'] = 'enum'
        elif leaf_info['type'] == "boolean":
            dic['type'] = 'bool'
    except:
        print(leaf_info)
    return dic

def make_simple_type_argument(leaf_info):
    """Generate the leaf-node's argument_spec.
    Args:
        leaf_infos: The result of get_leafinfos_from_xml_dict().
    Returns:
        A dict.The leaf-node's argument_spec.
        The leaf-node's argument_spec has (type, default, required) attribution which come from leaf_info.
    """
    leaf_args = {}
    if not leaf_info:
        return leaf_args
    if leaf_info['type'] == "enumeration":
        leaf_args["choices"] = leaf_info["choices"]
    else:
        leaf_args["type"] = leaf_info["type"]
        if leaf_args["type"] == "boolean":
            leaf_args["type"] = "bool"
        if leaf_args["type"] == "string":
            leaf_args["type"] = "str"

    if leaf_info["default"]:
        leaf_args["default"] = leaf_info["default"]

    leaf_args["required"] = leaf_info["required"]

    return leaf_args


# def make_arg_from_xpath(xpath, args):
#     """Add the xpath-key into argument_spec.
#     Args:
#         xpath: The xpath-key.
#         args: The argument_spec.
#     Returns:
#         A dict. The argument_spec of get.
#     """
#     sub_arg = args
#     nodes = xpath.split("/")
#     sub_arg["options"].update(dict(all=dict(type="bool", default=False)))
#     argument_spec = sub_arg = {nodes[-1]: sub_arg}
#     nodes = nodes[-2:0:-1]
#
#     for node in nodes:
#         sub_arg.update(dict(all=dict(type="bool", default=False)))
#         argument_spec = {node: dict(type="dict", options=sub_arg)}
#         sub_arg = argument_spec
#     return argument_spec

def make_argument_spec(xml_dict, yang_parser, xmlns_info, key_list, list_list):
    """Generate the argument_spec.
    Args:
        xml_dict: An OrderedDict. Recording the xml's structure.
        yang_parser: The yang_handler.
        xml_file:  The path of xml file.
    Returns:
        A dict.
        The output is for argument_spec.
        argument spec which is initial argument for class AnsibleModule
        Since netconf package for "get" and "config" support different operation scope. So for "get" package
        we will add two extra var to indicat:
            1. all: set True if we want get all data for while container/list, or we want get all instance of a leaf
            2. value: set it if we want get special instance of a leaf which's value equal to value we set
        And for "config" package we should make all leafs can be config no matter which level it is.
    """
    def _make_config_args(xpath, leaf_infos, list_list):
        """Use leaf_infos to generate config's argument_spec.(recursive function).
        Args:
            leaf_infos: The result of get_leafinfos_from_xml_dict().
        Returns:
            A dict.The config's argument_spec
        """
        argument_spec = OrderedDict()
        if isinstance(leaf_infos, OrderedDict):
            for node_name in leaf_infos:
                xpath = xpath + "/" + node_name
                complex_node = False
                for val in leaf_infos[node_name].values():
                    if isinstance(val, dict):
                        complex_node = True
                        break
                if not complex_node:
                    single_arg = make_simple_type_argument(
                        leaf_infos[node_name])
                    if not single_arg:
                        logging.warning(
                            "node %s has no defination in yang files", node_name)
                        continue
                    else:
                        argument_spec[node_name] = single_arg
                else:
                    node_spec = _make_config_args(xpath, leaf_infos[node_name], list_list)
                    list_flag = False
                    for sub_node_name in leaf_infos[node_name].keys():
                        temp_xpath = xpath + "/" + sub_node_name
                        if temp_xpath in list_list:
                            list_flag = True
                            break
                    if list_flag:
                        argument_spec[node_name] = dict(
                            type="list", elements="dict", options=node_spec)
                    else:
                        argument_spec[node_name] = dict(
                            type="dict", options=node_spec)
                xpath = xpath[:-(len(node_name) + 1)]

        return argument_spec

    def _make_get_args(xpath, leaf_infos, key_list, list_list):
        """Use leaf_infos to generate get's argument_spec.(recursive function).
        Args:
            xpath: The xpath-key.
            leaf_infos: The result of get_leafinfos_from_xml_dict().
        Returns:
            A dict. The argument_spec of get.
        """
        def _make_args_from_var_nodes(xpath, var_nodes, key_list, list_list):
            """Use leaf_infos to generate get's argument_spec.
            Args:
                var_nodes: The result of get_leafinfos_from_xml_dict().
            Returns:
                A dict. The get's argument_spec without xpath-key.
            """
            leafs = OrderedDict()
            leafs["options"] = OrderedDict()
            if not isinstance(var_nodes, OrderedDict):
                return None
            for key, v in var_nodes.items():
                xpath = xpath + "/" + key
                if isinstance(v, OrderedDict):
                    sub_arg = _make_args_from_var_nodes(xpath, v, key_list, list_list)
                    if xpath in list_list:
                        sub_arg["options"].update(
                            get_all=dict(type="bool", defalut=False))
                        leafs["options"].update({key: sub_arg})
                        leafs["type"] = "list"
                        leafs["elements"] = "dict"
                    else:
                        sub_arg["options"].update(
                            get_all=dict(type="bool", defalut=False))
                        leafs["options"].update({key: sub_arg})
                        leafs["type"] = "dict"
                else:
                    leaf_args = {'type': 'dict', "options": {
                        "get_all": dict(type="bool", defalut=False), "get_value": dict()}}
                    if v.get("required") and xpath not in key_list:
                        v["required"] = False
                    leaf_args["options"]["get_value"].update(make_simple_type_argument(v))
                    leafs["options"].update({key: leaf_args})
                    if xpath in list_list:
                        leafs["type"] = "list"
                        leafs["elements"] = "dict"
                    else:
                        leafs["type"] = "dict"
                xpath = xpath[:-(len(key) + 1)]
            return leafs

        argument_spec = _make_args_from_var_nodes(xpath, leaf_infos, key_list, list_list)

        return argument_spec["options"]

    schema_spec = None
    # Use leaf_info to generate argument_spec.
    leaf_infos = get_leafinfos_from_xml_dict(xml_dict, yang_parser, xmlns_info)
    for oper in xml_dict.keys():
        xpath = ""
        if oper == "config" or oper == "rpc":
            schema_spec = _make_config_args(xpath, leaf_infos, list_list)
        else:    # oper == "get" or oper == "get-config"
            schema_spec = _make_get_args(xpath, leaf_infos, key_list, list_list)

    return schema_spec


#!/usr/bin/env python
# -*- coding: utf-8

import os
import logging
from collections import OrderedDict

def get_xml_feature_name(file):

    return os.path.basename(file).split('_')[0]

def get_config_xml_head(head_xpath):
    """Get xml head string when config.
    Args:
        head_xpath: The string of xpath_key.
    Returns:
        xml_head_str: The xml head str.
    """
    xml_head_str = "<config>"
    for item in head_xpath.split("/")[1:]:
        xml_head_str = xml_head_str + "<" + item + ">"
    return xml_head_str


def get_rpc_head(xml_dict):
    """Get the rpc head.
    Args:
        xml_dict: An OrderedDict. Recording the xml's structure.
    Returns:
        xml_head: The xml head str.
        feature_name: The child node name of config|filter.
        xml_type: The type of xml.
        if xml_type is config,generate the xml_head with xpath_key.
        if xml_type is filter,the xml_head is '<filter type="subtree">'.
    """
    xml_head = ""
    feature_name_list = []
    xml_type = ""
    try:
        for key, value in xml_dict.items():
            xml_type = key
            if (value is None):
                continue
            for feature_key in value.keys():
                feature_name_list.append(feature_key)
        if xml_type == "config":
            xml_head = '<config>'
        elif xml_type == "rpc":
            xml_head = ""
        else:
            xml_head = '<filter type="subtree">'
    except Exception:
        logging.error("fun get_rpc_head() run error.")

    return [xml_head, feature_name_list, xml_type]

def get_config_xml_tail(tail_xpath):
    """Get xml tail string when config.
    Args:
        tail_xpath: The string of xpath_key.
    Returns:
        xml_tail_str: The xml tail str.
    """
    xml_tail_str = ""
    tail_xpath_list = tail_xpath.split("/")[1:]
    for item in reversed(tail_xpath_list):
        xml_tail_str = xml_tail_str + "</" + item + ">"
    xml_tail_str += "</config>"
    return xml_tail_str


def get_rpc_tail(xml_dict):
    """Get the rpc tail.
    Args:
        xml_dict: An OrderedDict. Recording the xml's structure.
    Returns:
        xml_tail: The xml tail str.
        if xml_type is config,generate the xml_tail with xpath_key.
        if xml_type is filter,the xml_tail is '</filter>'.
    """
    xml_tail = ""
    try:
        for key, value in xml_dict.items():
            xml_type = key
        if xml_type == "config":
            xml_tail = '</config>'
        elif xml_type == "rpc":
            xml_tail = ""
        else:
            xml_tail = '</filter>'
    except Exception:
        logging.error("fun get_rpc_tail() run error.")
    return xml_tail



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import logging
import os
import re


USER_CHECK_STMT = """class UserCheck(object):
    def __init__(self, params, infos):
        #  user configuration get from AnsibleModule().params
        self.params = params
        # leaf infos from yang files
        self.infos = infos

    # user defined check method need startswith "check_"
    # return 0 if not pass check logic, else 1
    def check_leaf_restrict(self):
        \"\"\"
            if leaf_1 configured, leaf2 shouble be configured
            and range shouble be in [10, 20]
        \"\"\"
        return 1   
"""


def get_imp_stmts(user_def_file):
    """Get 'import xxx' statement and 'from xxx import xxx' statement from userCheck.py.
    Args:
        user_def_file: The path of userCheck.py.
    Returns:
        xml_head_str: The import string of userCheck.py.
    """
    if not os.path.isfile(user_def_file):
        logging.error("%s is not exist or is not file", user_def_file)
        return

    logging.info("merge user check definitions from script %s", user_def_file)
    imp_decs = []
    idx = 0

    try:
        with open(user_def_file, encoding="utf-8") as file:
            lines = file.readlines()
            while idx < len(lines):
                line = lines[idx]
                # get 'import xxx' statement and 'from xxx import xxx' statement.
                if line.startswith("import ") or line.startswith("from "):
                    imp_dec = line
                    # The import have too many fun which need to change line.
                    while re.match(r".*\\\s+", line):
                        idx += 1
                        line = lines[idx]
                        imp_dec += line
                    imp_decs.append(imp_dec.rstrip())

                idx += 1
    except IOError as error:
        logging.error("can't process user define file %s, because %s", user_def_file, error)

    return "\n".join(imp_decs)


def get_user_check_cls_def(user_def_file):
    """Get 'class UserCheck(object):' statement from userCheck.py.
    Args:
        user_def_file: The path of userCheck.py.
    Returns:
        xml_head_str: The 'class UserCheck' statement of userCheck.py.
    """
    if not os.path.isfile(user_def_file):
        logging.error("%s is not exist or is not file", user_def_file)
        return

    logging.info("merge user check definitions from script %s", user_def_file)
    cls_str = "\n"
    is_cls_code = False
    idx = 0

    try:
        with open(user_def_file, encoding="utf-8") as file:
            lines = file.readlines()
            while idx < len(lines):
                line = lines[idx]
                # get code for class UserCheck
                if re.match(r"^class UserCheck\(object\):\s+$", line):
                    is_cls_code = True
                    cls_str += line
                    idx += 1
                    continue
                if is_cls_code:
                    if not re.match(r"\s+", line):
                        break
                    cls_str += line

                idx += 1
    except IOError as error:
        logging.error("can't process user define file %s, because %s", user_def_file, error)

    return cls_str



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re

DOCUMENTATION = """
description: get the ansible path.
"""


def get_ansible_path():
    '''
    get the ansible path
    :return:list
    '''
    res = os.popen('ansible --version').read()
    try:
        ansible_path = re.search(r'ansible python module location = (.*)', res).group(1)
    except Exception:
        raise Exception('No ansible_path!')
    return [ansible_path.strip() + '/modules/network/ne',
            ansible_path.strip() + '/module_utils/network/ne']


if __name__ == '__main__':
    a = get_ansible_path()
    print(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
import logging


def check_params(leaf_info, params, module):
    """Check all input params"""
    # Other parameters to be added to the synchronization to add
    # check
    return


def check_int(v_range, params, module):
    '''
    Check int type
    :param range:
    :param params:
    :param module:
    :return:
    '''
    if params and isinstance(params, dict):
        params = params["value"]
    if not params:
        return
    for long in v_range:
        if int(params) < int(long[0]) or int(params) > int(long[1]):
            logging.error('Error: %s not in the range from %s to %s.' % (params, long[0], long[1]))
            module.fail_json(msg='Error: %s not in the range from %s to %s.' % (params, long[0], long[1]))


def check_string(length, pattern, params, module):
    '''
    Check string type
    :param length:
    :param pattern:
    :param params:
    :param module:
    :return:
    '''
    if params and isinstance(params, dict):
        params = params["value"]
    if not params:
        return
    for long_str in length:
        if len(params) > long_str[1] or len(params.replace(' ', '')) < long_str[0]:
            logging.error('Error:%s is not in the range from %d to %d.' % (params, long_str[0], long_str[1]))
            module.fail_json(
                msg='Error:%s is not in the range from %d to %d.' % (params, long_str[0], long_str[1]))
    if pattern:
        for sub_re in pattern:
            if not re.match(sub_re, params):
                logging.error('Error:The input %s format is incorrect' % params)
                module.fail_json(msg='Error:The input %s format is incorrect' % params)



import os
import sys
import logging
import traceback
if sys.version < '3':
    import Queue


START_ERROR = "\n##########ANSIBLE_GEN_ERROR_START##########\n"
END_ERROR = "\n###########ANSIBLE_GEN_ERROR_END###########"
START_OPERATION_WARNING = "\n##########USER_OPERATION_WARNING_START##########\n"
END_OPERATION_WARNING = "\n###########USER_OPERATION_WARNING_END###########"
OLD_PROCESS_MESSAGE_LEN = 0


def error_write(error_string):
    """Write the error message into log file.
    Args:
        error_string: The error message.
    """
    exc_info = sys.exc_info()
    exc_value = exc_info[1]
    if sys.version < '3':
        if type(exc_value) == Queue.Empty:
            exc_value = None

    if exc_value and traceback.format_exc().strip() != 'None':
        logging.error(traceback.format_exc())
    logging.error("%s%s%s" % (START_ERROR, error_string, END_ERROR))


def operation_warning_write(operation_warning_string):
    """Write the warning message into log file.
    Args:
        operation_warning_string: The warning message.
    """
    logging.error("%s%s%s" % (START_OPERATION_WARNING, operation_warning_string, END_OPERATION_WARNING))


def xml_structure_except(exception_object, file_path):
    file_name = os.path.basename(file_path)
    syntax_msg_str = ""
    for item in exception_object.args:
        syntax_msg_str = "\n " + syntax_msg_str + str(item)
    if not file_path.endswith("_example.xml"):
        script_name = ""
        if file_path.split(os.sep)[-2]:
            script_name = file_path.split(os.sep)[-2]
        error_msg = "Script of %s.py can not be generated,because the structure of xml is incorrect." \
                  "\nPlease check the file %s:" % (script_name, file_path) \
                  + syntax_msg_str
        error_write(error_msg)
    else:
        warning_msg = "EXAMPLE of %s can not be generated,because the structure of xml is incorrect." \
                  "\nPlease check the file %s:" % (file_name, file_path) \
                  + syntax_msg_str
        operation_warning_write(warning_msg)


def print_progress_bar(completed_progress, process_message):
    """Print current progress and prompt message.
    Args:
        completed_progress: The value of current progress.
        process_message: The prompt message.
    """
    global OLD_PROCESS_MESSAGE_LEN
    space_len = 0
    if OLD_PROCESS_MESSAGE_LEN > len(process_message):
        space_len = OLD_PROCESS_MESSAGE_LEN - len(process_message)
    process_bar = "Process:{0}% |{1}|".format(str(completed_progress),
                                              str(int(completed_progress) * '=' + int(100 - completed_progress) * '-'))
    sys.stdout.write(u'\r')
    if sys.version < '3':
        sys.stdout.write(process_bar.decode("utf-8") + process_message.decode('utf-8') + str(space_len * ' '))
    else:
        sys.stdout.write(process_bar + process_message + str(space_len * ' '))
    sys.stdout.flush()
    OLD_PROCESS_MESSAGE_LEN = len(process_message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
from collections import OrderedDict
from .utils.base_util import operation_warning_write
from .utils.xml_parse.xml_parser import is_sub_xml, xml_to_ordered_dict
from .utils.xml_parse.xml_parser import get_operations_attr

if sys.version < '3':
    str = unicode

EXAMPLE_STR = ''
LEVEL = 0
module_files = []

def get_module_file(xml_dir, mod_files):
    """Get instance-xml's path.
    Args:
        xml_dir: The path of full-xml.
    Returns:
        A list. Saving paths of instance-xml.
    """
    xml_files = os.listdir(xml_dir)
    for xml_file in xml_files:
        path = os.path.join(xml_dir, xml_file)
        if os.path.isdir(path):
            get_module_file(path, mod_files)
        else:
            if os.path.isfile(path):
                if xml_file.endswith("_example.xml"):
                    mod_files.append(path)
                elif len(xml_files) == 1:
                    mod_files.append(path)
    return mod_files


def gen_dict(xml_file, full_xml_file):
    """
    Get xml_structure info and operation info.
    """
    xml_filename = xml_file.split('.')[0].split('/')[-1]
    if is_sub_xml(xml_file, full_xml_file):
        xml_dict = xml_to_ordered_dict(xml_file)
        operation_dict = get_operations_attr(xml_file)
        return [{xml_filename: xml_dict}, {xml_filename: operation_dict}]
    else:
        operation_error_str = "EXAMPLE of instance-xml can not be generated," \
                    "because the instance-xml is not a substructure of full-xml,please check:" \
                    "\nthe path of instance-xml: %s." \
                    "\nthe path of full-xml: %s." % (xml_file, full_xml_file)
        operation_warning_write(operation_error_str)
        return [{}, {}]


def gen_examples(module_files_dict, script_name, pkg_type, operation_dict_result, argument_spec):
    """
    Generate example string.
    """
    global EXAMPLE_STR
    global LEVEL
    if pkg_type == "config":
        process = get_cfg_str
    elif pkg_type == "rpc":
        process = get_cfg_str
    else:
        process = get_filter_str
    examples_str = '---' + "\n"
    examples_str += '- name: %s' % script_name + '\n'
    examples_str += '  hosts: ne_test' + '\n'
    examples_str += '  connection: netconf' + '\n'
    examples_str += '  gather_facts: no' + '\n'
    examples_str += '  vars:' + '\n'
    examples_str += '    netconf:' + '\n'
    examples_str += '      host: "{{ inventory_hostname }}"' + '\n'
    examples_str += '      port: "{{ ansible_ssh_port }}"' + '\n'
    examples_str += '      username: "{{ ansible_user }}"' + '\n'
    examples_str += '      password: "{{ ansible_ssh_pass }}"' + '\n'
    examples_str += '      transport: netconf' + '\n'
    examples_str += '\n'
    examples_str += '  tasks:\n'
    examples_str += '\n'
    for file_dict in module_files_dict:
        for key, v in file_dict.items():
            examples_str += '  - name: ' + os.path.basename(key) + '\n'
            examples_str += '    %s:' % script_name + '\n'
            for xml_type, xml_dict in v.items():
                examples_str += '      operation_type: ' + xml_type + '\n'
                if xml_type == 'config':
                    # generate operation_specs
                    if operation_dict_result:
                        for operation_dict in operation_dict_result:
                            for operation_key, operation_value in operation_dict.items():
                                if operation_value:
                                    if operation_key == key:
                                        examples_str += '      operation_specs: ' + '\n'
                                        for xpath, value in operation_value.items():
                                            examples_str += '        - path: ' + xpath + '\n'
                                            examples_str += '          operation: ' + value + '\n'
                                        break

                if pkg_type == "config":
                    examples_str += process(xml_dict, argument_spec)
                else:
                    examples_str += process(xml_dict, argument_spec)
            EXAMPLE_STR = ''
            LEVEL = 0

        examples_str += '      provider: "{{ netconf }}"' + '\n'
        examples_str += '\n'
    return examples_str

def get_filter_str(data, argument_spec):
    """
    Generate the example of filter.
    """
    def get_leaf_str(data, indent, argument_spec, parent_node_type=None):
        leaf_str = ''
        list_flag = ""
        indent_str = "  "
        if parent_node_type and parent_node_type == "list":
            list_flag = "- "
            indent_str = "    "
        for k, v in data.items():
            if 'options' not in argument_spec[k]:
                print(argument_spec[k])
            current_argument_spec = argument_spec[k].get("options")
            parent_node_type = argument_spec[k].get("type")
            if v is None and argument_spec[k].get("type") == "list":
                container_end_flag = "- "
            else:
                container_end_flag = ""
            if not isinstance(v, list):
                v = [v]
            for node in v:
                if not isinstance(node, OrderedDict):
                    leaf_str += "".join(indent) + list_flag + k + ': \n'
                    indent.append(indent_str)
                    node_str = container_end_flag + 'get_all: True'
                    if node:
                        node_str = 'get_value: ' + node
                        if current_argument_spec.get("get_value") and current_argument_spec["get_value"].get("type") == "str":
                            node_str = 'get_value: \"' + node + '\"'
                    leaf_str += "".join(indent) + node_str + '\n'
                    indent.pop()
                else:
                    leaf_str += "".join(indent) + list_flag + k + ': ' + '\n'
                    indent.append(indent_str)
                    leaf_str += get_leaf_str(node, indent, current_argument_spec, parent_node_type)
                    indent.pop()
        return leaf_str

    indent = ["  ", "  ", "  "]
    task_str = ""
    if data:
        task_str += get_leaf_str(data, indent, argument_spec)
    else:
        task_str += '    ' + '  ' * LEVEL + 'get_all: True' + '\n'
    return task_str


def get_cfg_str(data, argument_spec, parent_node_type=None):
    """
    Generate the example of config.
    """
    if isinstance(data, OrderedDict):
        global LEVEL
        LEVEL += 1
        for k, v in data.items():
            global EXAMPLE_STR
            if not argument_spec.get(k):
                continue
            if argument_spec[k].get("options"):
                argument_spec_options = argument_spec[k]["options"]
            else:
                argument_spec_options = argument_spec
            if argument_spec[k].get("type"):
                parent_node_type_temp = argument_spec[k]["type"]
            else:
                parent_node_type_temp = None

            if isinstance(v, OrderedDict):
                if parent_node_type and parent_node_type == "list":
                    EXAMPLE_STR += '    ' + '  ' * LEVEL + '- ' + k + ': ' + '\n'
                    LEVEL += 1
                    get_cfg_str(v, argument_spec_options, parent_node_type_temp)
                    LEVEL -= 1
                else:
                    EXAMPLE_STR += '    ' + '  ' * LEVEL + k + ': ' + '\n'
                    get_cfg_str(v, argument_spec_options, parent_node_type_temp)
            elif isinstance(v, list):
                for item in v:
                    if parent_node_type and parent_node_type == "list":
                        EXAMPLE_STR += '    ' + '  ' * LEVEL + '- ' + k + ': ' + '\n'
                        LEVEL += 1
                        get_cfg_str(item, argument_spec_options, parent_node_type_temp)
                        LEVEL -= 1
                    else:
                        EXAMPLE_STR += '    ' + '  ' * LEVEL + k + ': ' + '\n'
                        get_cfg_str(item, argument_spec_options, parent_node_type_temp)
            else:
                if v:
                    if argument_spec[k].get("type") and argument_spec[k]["type"] == "str":
                        v = '\"' + v + '\"'
                    EXAMPLE_STR += '    ' + '  ' * LEVEL + k + ': ' + v + '\n'
                else:
                    if argument_spec[k].get("type") and argument_spec[k]["type"] == "dict":
                        value_empty = '{}'
                    elif argument_spec[k].get("type") and argument_spec[k]["type"] == "list":
                        value_empty = '[]'
                    else:
                        value_empty = '\"\"'
                    EXAMPLE_STR += '    ' + '  ' * LEVEL + k + ': ' + value_empty + '\n'

    LEVEL -= 1
    return EXAMPLE_STR


def create_example(full_xml_file_path, script_name, pkg_type, argument_spec):
    """Generate example.
    Args:
        full_xml_file_path: The path of full-xml.
        script_name: The file name of full-xml.
        pkg_type: Type of xml. The value is "config" or "filter".
    Returns:
        example_str: The example string.
        Compare the instance-xml's structure with full-xml,to make sure instance-xml is a subset of full-xml.
    """
    # Get all instance-xml.
    module_files = get_module_file(os.path.dirname(full_xml_file_path), [])
    module_files.sort()
    module_files_dict = []
    operation_dict_result = []
    for file in module_files:
        xml_res = gen_dict(file, full_xml_file_path)
        module_files_dict.append(xml_res[0])
        operation_dict_result.append(xml_res[1])
    example_str = gen_examples(module_files_dict, script_name, pkg_type, operation_dict_result, argument_spec)

    return example_str



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
import os
import sys
import copy
import logging
import json
from collections import OrderedDict
import hashlib
import xmltodict
from lxml import etree
from xml.dom.minidom import parseString
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.connection import Connection, ConnectionError
from ansible.module_utils.network.ne.ne import get_nc_config, ne_argument_spec, get_nc_connection, to_text, to_string, execute_nc_action_yang
from ansible.module_utils.network.ne.common_module.checkparams import check_params
from ansible.module_utils.network.ne.common_module import xmltodict
from ansible.module_utils.network.ne.common_module.xml_build_with_xmlns import xml_parser_join_xmlns

try:
    from ncclient.xml_ import to_xml
    HAS_NCCLIENT = True
except ImportError:
    HAS_NCCLIENT = False

try:
    import xml.etree.cElementTree as ET
except:
    import xml.etree.ElementTree as ET

try:
    from mediator.netconf_translate import translate_edit_config_content, translate_query_filter_content
    HAS_MEDIATOR = True
except ImportError:
    HAS_MEDIATOR = False

class ConfigBase(object):
    """Create a ConfigBase class object"""

    def __init__(self, *args):
        self.argument_spec, self.leaf_info, self.namespaces, self.business_tag, \
        self.xml_head, self.xml_tail, self.key_list = args
        self.module = None

        # config result state
        self.changed = False
        self.updates_cmd = list()
        self.results = dict()
        self.translate_ietf = str()
        self.ietf_routing = str()
        self.proposed = dict()
        self.existing = dict()
        self.end_state = dict()

    def init_module(self):
        """ init module """
        self.module = AnsibleModule(
            argument_spec=self.argument_spec,
            supports_check_mode=True)
        return self.module

    def get_operation_type(self):
        """Get operation type"""
        self.init_module()
        return self.module.params

    # config_set Echo check function
    def check_response(self, xml_str):
        """Check if response message is already succeed."""
        if "<ok/>" not in xml_str:
            self.module.fail_json(msg=xml_str)
        elif " <rpc-error>" in xml_str:
            self.module.fail_json(msg=xml_str)

    def get_body_xml(self, oper):
        """Config_set construct the body part of the message."""
        params_copy = copy.deepcopy(self.module.params)
        business_params = {}
        for business in self.business_tag:
            business_params[business] = params_copy[business]
        if oper == 'config':
            new_params = {'root': business_params}
        else:
            # Remove the non-key node, otherwise the get message is incorrect.
            new_business_params = self.remove_not_key(business_params, self.key_list, "")
            new_params = {'root': new_business_params}
        body_xml = parseString(self.params_to_xml(oper, new_params)).toprettyxml()
        body_xml_list = re.compile(r"<root>(.*?)</root>", re.S).findall(body_xml)
        if not body_xml_list:
            return
        body_xml_str = body_xml_list[0]
        return body_xml_str

    def remove_not_key(self, dic, paths, spath):
        """Remove non-key nodes."""
        for k in list(dic.keys()):
            spath = spath + "/" + k
            if isinstance(dic[k], dict):
                dic[k].update(
                    self.remove_not_key(dic[k], paths, spath))
            elif isinstance(dic[k], list):
                temp_list = []
                for item in dic[k]:
                    temp_list.append(
                        self.remove_not_key(item, paths, spath))
                dic.update({k: temp_list})
            else:
                if spath in paths:
                    pass
                else:
                    if dic[k] is None:
                        del dic[k]
                    else:
                        dic[k] = None
            spath = spath[:-(len(spath.split("/")[-1]) + 1)]
        return dic

    def str_to_xml(self, cfg_str):
        """Convert the spliced string to xml."""
        cfg_str_to_doc = parseString(cfg_str).toprettyxml()
        xml_list = [line.strip() for line in cfg_str_to_doc.split("\n") if line]
        update_xml = parseString(''.join(xml_list)).toprettyxml()
        root = etree.fromstring(update_xml)
        if self.module.params.get('operation_specs', None):
            update_xml = self.create_operation_xml(self.module.params, root)
        return self.create_operation_namespace(update_xml)

    def create_operation_xml(self, node_params, root):
        """Create operation xml"""
        # Add a user-defined operation
        update_xml = ""
        for dic in node_params.get('operation_specs'):
            path = dic['path']
            operation = dic['operation']
            user_tag = root.xpath(path)
            user_tag[0].set("operation", operation)
            update_xml = parseString(etree.tostring(root)).toprettyxml(indent='\t', newl='\n')
        return update_xml

    def create_operation_namespace(self, update_xml):
        """Add a namespace to the xml file"""
        update_xml_temp = xml_parser_join_xmlns(update_xml, self.namespaces, "config")
        ret_str = update_xml_temp.replace(" operation", " nc:operation")
        return ret_str

    def config_str(self):
        """Config_set message stitching"""
        cfg_str = ''
        cfg_str += self.xml_head.replace(' nc:operation', ' operation')
        cfg_str += self.get_body_xml('config')
        cfg_str += self.xml_tail
        return cfg_str

    def load_json(self, xml):
        """json to xml"""
        try:
            json_str = xmltodict.parse(xml)
            json_s = json.dumps(json_str, indent=4)
            return json.loads(json_s)
        except Exception:
            print("xmltojson failed!")

    def json_to_xml(self, json):
        """xml to json"""
        xml_str = xmltodict.unparse(json, pretty='True')
        return xml_str

    def netconf_set_config(self):
        """The final config_set message is sent to the controlled machine."""
        cfg_str = self.config_str()
        xml_str = self.str_to_xml(cfg_str)
        ietf_xml_json = self.load_json(xml_str)
        self.ietf_routing = self.json_to_xml(ietf_xml_json)
        if HAS_MEDIATOR:
            xml_cfg_str = translate_edit_config_content(xml_str).replace('True', 'true'). \
                replace('False', 'false')
        else:
            xml_cfg_str = xml_str
        self.translate_ietf = xml_cfg_str
        recv_xml = self.set_nc_config_without_default_operation(self.module, xml_cfg_str)
        self.check_response(recv_xml)
        return recv_xml

    # ietf config message not need the param:<default-operation>
    # Duplicate the fun:set_nc_config
    def set_nc_config_without_default_operation(self, module, xml_str):
        """ set_config """
        conn = get_nc_connection(module)
        if xml_str is not None:
            try:
                out = conn.edit_config(target='running', config=xml_str, error_option='rollback-on-error')
            except ConnectionError as exc:
                module.fail_json(msg=to_text(exc))
            finally:
                pass
        else:
            return None
        return to_string(to_xml(out))

    def get_config_str(self):
        """config_get Message stitching"""
        get_cfg_str = ''
        replace_xml_head = self.xml_head.replace('config', 'filter type="subtree"')
        if re.compile(r' nc:operation="\w*"?'):
            new_xml_head = re.compile(r' nc:operation="\w*"?').sub('', replace_xml_head)
        else:
            new_xml_head = replace_xml_head
        get_cfg_str += new_xml_head
        get_cfg_str += self.get_body_xml('get')
        get_cfg_str += self.xml_tail.replace('config', 'filter')
        return get_cfg_str

    def params_to_xml(self, oper, params):
        """Convert params dictionary type to xml format."""
        return self.to_xml(oper, params)

    def to_xml(self, oper, params):
        """Deep traversal of the dictionary and remove the default value."""
        xml_list = []
        for key, value in params.items():
            if isinstance(value, dict):
                xml_value = self.to_xml(oper, value)
                xml_str = self.xml_tag(key, xml_value)
            elif isinstance(value, list):
                xml_value = ""
                for item in value:
                    xml_value = xml_value + self.to_xml(oper, item)
                xml_str = self.xml_tag(key, xml_value)
            else:  # The type of value might be None/int/str/bool
                if value is None and oper == "config":
                    continue
                if type(value) == bool:
                    value = str(value).lower()
                xml_str = self.xml_tag(key, value)
            xml_list.append(xml_str)
        return ''.join(xml_list)

    def xml_tag(self, key, value):
        """Generate xml format for each line"""
        xml_str = ("<%s>{value}</%s>" % (key, key)) \
            .format(key=key, value=value if value is not None else '')
        return xml_str

    def get_xml_str(self):
        """Get the xml which is get message."""
        get_cfg_str = self.get_config_str()
        cfg_str_to_doc = parseString(get_cfg_str).toxml()
        xml_list = [line.strip() for line in cfg_str_to_doc.split("\n") if line]
        get_xml = parseString(''.join(xml_list)).toprettyxml()
        get_xml_str = get_xml.replace('<?xml version="1.0" ?>', '')
        xml_str = parseString(get_xml_str).toprettyxml()
        return xml_str

    def netconf_get_config(self):
        """The final Config_get message is sent to the controlled machine."""
        xml_str = self.get_xml_str()
        update_xml_result = xml_parser_join_xmlns(xml_str, self.namespaces, 'filter')
        get_str = update_xml_result.replace('<?xml version="1.0" ?>', '')
        if HAS_MEDIATOR:
            translate_xml_str = translate_query_filter_content(get_str)
        else:
            translate_xml_str = get_str
        return self.get_info_process(translate_xml_str)

    # The config_get packet is sent to return the current configuration parameters of the device.
    def get_info_process(self, xml_str):
        """Get current dldp existed configuration"""
        conf = dict()
        # Send a Get message
        # Parsing 1: delete the useless string, pay attention to the replacement according to the business
        con_obj = get_nc_config(self.module, xml_str)

        # Parsing 2: No data detection
        if "<data/>" in con_obj:
            return conf

        # Parsing 3: Extract all nodes in the root directory
        new_con_obj_temp = con_obj.replace('\r', '').replace('\n', ''). \
            replace('<?xml version=\"1.0\" encoding=\"UTF-8\"?>', "")
        new_con_obj = re.sub(r'\sxmlns(:.*?)=".*?"|\sxmlns=".*?"', "", new_con_obj_temp)
        xml_to_dict = xmltodict.parse(new_con_obj)
        conf = xml_to_dict["data"]
        return conf

    def convert_to_set(self, d, xpath=''):
        '''
        Turn a dictionary into a collection
        :param d:
        :param xpath:
        :return:
        '''
        res = set()
        for key, val in d.items():
            if isinstance(val, dict):
                res = res | self.convert_to_set(val, xpath + "/" + key)
            elif isinstance(val, list):
                res.add((xpath + "/" + key, str(val)))
            else:
                res.add((xpath + "/" + key, val))
        return res

    def convert_to_dict(self, diff_s, dep, prefix):
        '''
        Turn the collection into a dictionary
        :param diff_s:
        :param dep:
        :param prefix:
        :return:
        '''
        if not diff_s:
            return {}
        res = dict()
        tmp_diff = {val for val in diff_s}
        for xpath, val in tmp_diff:
            if xpath.count("/") == dep:
                leaf_name = xpath[xpath.rfind("/") + 1:]
                prefix.update([xpath])
                res[leaf_name] = val
                diff_s.remove((xpath, val))
        tmp_diff = {val for val in diff_s}
        for xpath, val in tmp_diff:
            if xpath not in prefix:
                res.update({xpath.split("/")[dep]: self.convert_to_dict(diff_s, dep + 1, prefix)})
        return res

    # Compare the differences between the two dictionaries
    def compare_two_dict(self, existing, end_state):
        """
        # 1. Convert two dictionaries into a collection
        # 2. Take the intersection of the two sets
        # 3. Take the difference between end_state and intersection as update_dic
        # 4. Determine if update_dic is empty, if there is a value, changed to true
        """
        existing_set = self.convert_to_set(existing)
        end_state_set = self.convert_to_set(end_state)
        base = existing_set & end_state_set
        diff_s = end_state_set - base or existing_set - base
        update_dic = self.convert_to_dict(diff_s, 1, set())
        if update_dic:
            self.changed = True
        return dict(update_dic)

    def get_existing(self):
        """
        Get existing configuration parameters
        :return:
        """
        self.existing = self.netconf_get_config()

    # Get the end_state parameters after execution
    def get_end_state(self):

        self.netconf_set_config()
        self.end_state = self.netconf_get_config()

    def get_update_cmd(self):
        """Get update_cmd parameters"""
        self.updates_cmd.append(
            self.compare_two_dict(self.existing, self.end_state))

    # Data returned to the user
    def show_result(self):
        """Show result"""
        # self.results['huawei'] = self.module.jsonify(self.translate_ietf)
        # self.results['ietf-routing'] = self.module.jsonify(self.ietf_routing)
        self.results['changed'] = self.changed
        self.results['end_state'] = self.end_state
        self.results['proposed'] = self.proposed

        self.results['existing'] = self.existing
        if self.changed:
          self.results['updates'] = self.updates_cmd
        else:
         self.results['updates'] = list()

        self.module.exit_json(**self.results)
        # self.module.exit_json(msg='ietf\n'+ self.ietf_routing+'\n\nhuawei\n'+self.translate_ietf)

    def run(self):
        """worker"""
        # module input info
        self.init_module()
        check_params(self.leaf_info, self.module.params, self.module)

        # return results
        #self.get_proposed()
        self.get_existing()
        self.get_end_state()
        self.get_update_cmd()
        self.show_result()

class GetBase(object):
    """Create a GetBase class object"""

    def __init__(self, *args):
        self.argument_spec, self.leaf_info, self.namespaces, self.business_tag, self.xml_head, self.xml_tail,\
        self.key_list = args
        self.module = None

        # filter result state
        self.changed = False
        self.proposed = dict()
        self.results = dict()
        self.end_state = dict()

    def init_module(self):
        """ init module """
        self.module = AnsibleModule(
            argument_spec=self.argument_spec,
            supports_check_mode=True)
        return self.module

    def get_body_xml(self):
        """Construct the body part of the message."""
        params_copy = copy.deepcopy(self.module.params)
        business_params = {}
        for business in self.business_tag:
            business_params[business] = params_copy[business]
        new_params = {'root': business_params}
        
        body_xml_temp = parseString(self.params_to_xml(new_params)).toprettyxml()
        body_xml = body_xml_temp.replace('<?xml version="1.0" ?>', '')
        body_xml_list = re.compile(r"<root>(.*?)</root>", re.S).findall(body_xml)
        if not body_xml_list:
            return
        body_xml_str = body_xml_list[0]
        return body_xml_str

    def get_config_str(self):
        """Filter_get message stitching."""
        get_cfg_str = ''
        get_cfg_str += self.xml_head
        get_cfg_str += self.get_body_xml()
        get_cfg_str += self.xml_tail
        return get_cfg_str

    def params_to_xml(self, params):
        """Convert params dictionary type to xml format."""
        instance_xml_root = {"root": params['root']}
        if instance_xml_root:
            for root_tag in instance_xml_root:
                root_tree = ET.Element(root_tag)
                self.to_xml(instance_xml_root[root_tag], root_tree)
            xmlstr = ET.tostring(root_tree, method='xml')
        return xmlstr

    # generate the body xml(Deep traversal dictionary)
    def to_xml(self, root_tree_dict, root_tree):
        """
        :param root_tree_dict:  the root of the data structure.
        :param root_tree:  the generate result.
        """
        for sub_node_tag, sub_node_value in root_tree_dict.items():
            if sub_node_tag != "get_all" and sub_node_tag != "get_value":
                # node exist in full.xml not in instance xml,do nothing
                if sub_node_value is None:
                    continue
                # node need to generate in xml
                elif isinstance(sub_node_value, list):
                    sub_element = ET.SubElement(root_tree, sub_node_tag)
                    for item in sub_node_value:
                        self.to_xml(item, sub_element)
                elif isinstance(sub_node_value, dict):
                    sub_element = ET.SubElement(root_tree, sub_node_tag)
                    # get value
                    value = None
                    for item in sub_node_value:
                        if item == "get_value":
                            value = sub_node_value[item]   # value: 1.int  2.str  3.boolean  4.None
                            break
                    if value is None:
                        value = ''
                    # <nodeName>True<nodeName> ---> <nodeName>true<nodeName>
                    if isinstance(value, bool):
                        value = str(value).lower()
                    # set value
                    if value or value == 0:
                        sub_element.text = str(value)

                    # get all status
                    all_status = False
                    for item in sub_node_value:
                        if item == "get_all":
                            all_status = sub_node_value[item]   # all: 1.True  2.False  3.None
                            break
                    if all_status is None:
                        all_status = False

                    # recursive
                    if all_status:
                        continue
                    else:
                        self.to_xml(sub_node_value, sub_element)
            else:
                continue

    def get_xml_str(self):
        """Get the xml which is get message."""
        get_cfg_str = self.get_config_str()
        cfg_str_to_doc = parseString(get_cfg_str).toxml()
        xml_list = [line.strip() for line in cfg_str_to_doc.split("\n") if line]
        update_xml = parseString(''.join(xml_list)).toprettyxml()
        get_str_temp = update_xml.replace('<?xml version="1.0" ?>', ''). \
            replace('type="subtree"', '')
        return get_str_temp

    def netconf_get_config(self):
        """The final Filter_get message is sent to the controlled machine."""
        xml_str = self.get_xml_str()
        get_str = xml_parser_join_xmlns(xml_str, self.namespaces, "filter")
        if HAS_MEDIATOR:
            translate_cfg_get = translate_query_filter_content(get_str)
        else:
            translate_cfg_get = get_str
        return self.get_info_process(translate_cfg_get)

    def get_config(self, module, xml_str, *args, **kwargs):
        """ get_config """
        conn = get_nc_connection(module)
        if xml_str is not None:
            try:
                response = conn.get_config(source='running', filter=xml_str)
            except ConnectionError as exc:
                module.fail_json(msg=to_text(exc))
            finally:
                pass
        else:
            return None
        return to_string(to_xml(response))

    # The get message is sent, and the current configuration parameters of the device are returned.
    def get_info_process(self, xml_str):
        """Get current dldp existed configuration"""
        conf = dict()
        # Send a get message
        # Parsing 1: delete the useless string, pay attention to the replacement according to the business
        if self.module.params["operation_type"] == "get":
            con_obj = get_nc_config(self.module, xml_str)
        else:  # ["operation_type"] == "get-config"
            con_obj = self.get_config(self.module, xml_str)
        #  Parsing 2: No data detection
        if "<data/>" in con_obj:
            return conf
        # Parsing 3: Extracting the echoed message
        new_con_obj_temp = con_obj.replace('\r', '').replace('\n', ''). \
            replace('<?xml version=\"1.0\" encoding=\"UTF-8\"?>', "")
        new_con_obj = re.sub(r'\sxmlns(:.*?)=".*?"|\sxmlns=".*?"', "", new_con_obj_temp)
        xml_to_dict = xmltodict.parse(new_con_obj)
        conf = {"result": xml_to_dict}
        return conf

    # get end_state parameters
    def get_end_state(self):
        '''
        get the end state
        :return:
        '''
        self.end_state = self.netconf_get_config()

    # Data returned to the user
    def show_result(self):
        """Show result"""
        self.results['changed'] = self.changed
        self.results['end_state'] = self.end_state
        self.results['proposed'] = self.proposed

        self.module.exit_json(**self.results)

    def run(self):
        """worker"""
        # module input info
        self.init_module()
        check_params(self.leaf_info, self.module.params, self.module)

        # return results
        #self.get_proposed()
        self.get_end_state()
        self.show_result()

class InputBase(object):
    def __init__(self, *args):
        self.argument_spec, self.leaf_info, self.namespaces, self.business_tag, \
        self.xml_head, self.xml_tail, self.key_list = args
        self.module = None

        # config result state
        self.results = dict()

    def init_module(self):
        """ init module """
        self.module = AnsibleModule(
            argument_spec=self.argument_spec,
            supports_check_mode=True)
        return self.module

    def get_operation_type(self):
        """Get operation type"""
        self.init_module()
        return self.module.params

    def xml_tag(self, key, value):
        """Generate xml format for each line"""
        xml_str = ("<%s>{value}</%s>" % (key, key)) \
            .format(key=key, value=value if value is not None else '')
        return xml_str

    def to_xml(self, params):
        """Deep traversal of the dictionary and remove the default value."""
        xml_list = []
        for key, value in params.items():
            if isinstance(value, dict):
                xml_value = self.to_xml(value)
                xml_str = self.xml_tag(key, xml_value)
            elif isinstance(value, list):
                xml_value = ""
                for item in value:
                    xml_value = xml_value + self.to_xml(item)
                xml_str = self.xml_tag(key, xml_value)
            else:  # The type of value might be None/int/str/bool
                if value is None:
                    continue
                if type(value) == bool:
                    value = str(value).lower()
                xml_str = self.xml_tag(key, value)
            xml_list.append(xml_str)
        return ''.join(xml_list)

    def dict_order(self, business_params, leaf_info):
        temp_ordered_dict = OrderedDict()
        key_list = list(leaf_info.keys())
        for key in key_list:
            if isinstance(business_params, dict) and key in business_params.keys():
                temp_ordered_dict[key] = leaf_info[key]

        for key, value in business_params.items():
            if isinstance(value, dict):
                temp_ordered_dict[key] = self.dict_order(business_params[key], leaf_info[key])
            elif isinstance(value, list):
                temp_list = []
                for instance_dict in value:
                    temp_list.append(self.dict_order(instance_dict, leaf_info[key]))
                temp_ordered_dict[key] = temp_list
            else:
                temp_ordered_dict[key] = value
        return temp_ordered_dict

    def get_body_xml(self):
        """Config_set construct the body part of the message."""
        params_copy = copy.deepcopy(self.module.params)
        business_params = {}
        for business in self.business_tag:
            business_params[business] = params_copy[business]
        new_params = OrderedDict()
        new_params["root"] = self.dict_order(business_params, self.leaf_info)
        body_xml = parseString(self.to_xml(new_params)).toprettyxml()
        body_xml_list = re.compile(r"<root>(.*?)</root>", re.S).findall(body_xml)
        if not body_xml_list:
            return
        body_xml_str = body_xml_list[0]
        return body_xml_str

    # config_set Echo check function
    def check_response(self, xml_str):
        """Check if response message is already succeed."""
        conf = dict()
        if " <rpc-error>" not in xml_str:
            new_con_obj_temp = xml_str.replace('\r', '').replace('\n', ''). \
                replace('<?xml version=\"1.0\" encoding=\"UTF-8\"?>', "")
            new_con_obj = re.sub(r'\sxmlns(:.*?)=".*?"|\sxmlns=".*?"', "", new_con_obj_temp)
            xml_to_dict = xmltodict.parse(new_con_obj)
            conf = {"result": xml_to_dict}
        return conf

    def netconf_set_config(self, cfg_str):
        """The final config_set message is sent to the controlled machine."""
        cfg_str = "<rpc>" + cfg_str + "</rpc>"
        xml_str_with_xmlns = xml_parser_join_xmlns(cfg_str, self.namespaces, "rpc")
        send_xml_str = xml_str_with_xmlns.replace('<rpc>', "").replace('</rpc>', "")
        # Send a Get message
        # Parsing 1: delete the useless string, pay attention to the replacement according to the business
        con_obj = execute_nc_action_yang(self.module, send_xml_str)
        conf = self.check_response(con_obj)
        return conf

    # Data returned to the user
    def show_result(self):
        """Show result"""
        cfg_str = self.get_body_xml()
        xml_str = parseString(cfg_str).toprettyxml()
        self.results['send_xml'] = xmltodict.parse(xml_str)
        self.results['output'] = self.netconf_set_config(cfg_str)
        self.module.exit_json(**self.results)

    def run(self):
        """worker"""
        # module input info
        self.init_module()
        check_params(self.leaf_info, self.module.params, self.module)
        # return results
        self.show_result()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
import codecs
import traceback
import logging
import shutil
import math
from optparse import OptionParser
if sys.version_info[0] == 2:
    from adapter import get_argument_spec_documentation as get_parser
    from generator import var, ansible_auto_scripts as script_gen
    from generator.move_file import *
    from adapter.utils import base_util
else:
    from ansible_gen.adapter import get_argument_spec_documentation as get_parser
    from ansible_gen.generator import var, ansible_auto_scripts as script_gen
    from ansible_gen.generator.move_file import *
    from ansible_gen.adapter.utils import base_util


if not sys.version > '3':
    import ConfigParser as configparser

    reload(sys)
    sys.setdefaultencoding('utf-8')
else:
    import configparser


USAGE = "Dynamically generate Ansible modules from yang and xml files, then deploy Ansible module"
__version__ = "0.5.2"

SCRIPT_GEN_LOG_FILE = "ansible_gen.log"
BASE_ERROR = r"##########ANSIBLE_GEN_ERROR_START##########"
END_ERROR = r"###########ANSIBLE_GEN_ERROR_END###########"
START_WARNING = r"##########USER_OPERATION_WARNING_START##########"
END_WARNING = r"###########USER_OPERATION_WARNING_END###########"
GENERATE_SCRIPT_MESSAGE = ""
SCRIPT_DIR_GLOBAL = ""

def parse_error_and_exit(filename):
    """Extract the error information from "ansible_gen.log".
    Args:
        filename: The path of "ansible_gen.log".
    """
    ret_val = 0
    handle = None
    try:
        handle = codecs.open(filename, encoding="utf-8")
        log_context = handle.read()
        findre = re.compile(r"%s.*?%s" % (BASE_ERROR, END_ERROR), re.DOTALL)
        error_list = findre.findall(log_context)
        warning_obj = re.compile(r"%s.*?%s" % (START_WARNING, END_WARNING), re.DOTALL)
        warning_list = warning_obj.findall(log_context)
        if error_list:
            print("Ansible-gen Execute Failed.")
            ret_val = 1
            for error_str in set(error_list):
                print(error_str)
        else:
            print("Ansible-gen Execute Success.")
        if warning_list:
            for operation_warning_str in set(warning_list):
                print(operation_warning_str)
    except Exception:
        ret_val = 1
    finally:
        if handle is not None:
            handle.close()
        sys.exit(ret_val)


def fill_args_from_cfg_file(args, options):
    """Config the OptionParser instance by default.cfg file.
    Args:
        args: The list record arguments.
        options: The instance record user input.
    """
    user_cfg_file = args[0] if args else ""
    default_cfg_file = os.path.join("/etc/ansible-gen/default.cfg")
    if options.default is not None:
        cfg_file = default_cfg_file
    else:
        cfg_file = user_cfg_file if user_cfg_file else default_cfg_file
    if not os.path.isfile(cfg_file):
        return
    cfg_info = configparser.ConfigParser()
    cfg_info.read(cfg_file)

    for name in ["yang_dir", "xml_dir", "log_dir", "script_dir"]:
        if not getattr(options, name):
            if cfg_info.has_option("defaults", name):
                setattr(options, name, cfg_info.get("defaults", name))
    if cfg_info.has_option("defaults", "log_level"):
        setattr(options, "log_level", cfg_info.get("defaults", "log_level"))
    else:
        options.log_level = "DEBUG"


def prepare_parser():
    """Get the parser default value.
    Returns:
        parser: OptionParser instance.
    """
    # if user dosn't give andy command line options, add -h manually
    if not sys.argv[1:]:
        sys.argv.append("-h")

    parser = OptionParser(description=USAGE, epilog="Ansible-gen {}".format(__version__))
    parser.add_option("-v", "--version", dest="version", default='',
                      action='store_true', help="version number info for program")

    return parser


def parse_args():
    """Parse the user input.
    Returns:
        options: The instance record user input.
        args: The list record arguments.
    """
    parser = prepare_parser()

    parser.add_option("-y", "--yang_dir", dest="yang_dir", default='',
                      help="the directory of yang_files.")
    parser.add_option("-r", "--resource", dest="xml_dir", default='',
                      help="the directory of ansible api description xml files.")
    parser.add_option("-p", "--script", dest="script_dir", default='',
                      help="the directory of previous generated ansible module which may has user define"
                           " check implementation.")
    parser.add_option("-l", "--log", dest="log_dir", default=None,
                      help="the log directory, name of log is ansible_gen.log")
    parser.add_option("-o", "--output", dest="output_dir", default=None,
                      help="the output dir for generated ansible modules")
    parser.add_option("--default", dest="default", default='', action='store_true',
                      help="get parameters from default config file /etc/ansible-gen/default.cfg")

    (options, args) = parser.parse_args()

    if options.version:
        print("ansible-gen", __version__)
        sys.exit()

    fill_args_from_cfg_file(args, options)

    if options.log_dir is None:
        options.log_dir = os.path.join(os.getcwd(), 'logs')
    if not os.path.exists(options.log_dir):
        os.makedirs(options.log_dir)

    log_file_name = os.path.join(options.log_dir, SCRIPT_GEN_LOG_FILE)

    if os.path.isfile(log_file_name):
        os.remove(log_file_name)
    if hasattr(options, "log_level"):
        log_level = options.log_level
    else:
        log_level = "DEBUG"

    logging.basicConfig(filename=log_file_name,
                        format='%(asctime)s,%(levelname)s,%(filename)s,%(lineno)d:%(message)s',
                        level=log_level)

    return options, args


def env_parse():
    """Get user input from the command line.
    Returns:
        options: The instance record user input.
    """
    (options, args) = parse_args()

    for name in ["yang_dir", "xml_dir"]:
        if not os.path.isdir(getattr(options, name)):
            sys.stderr.write("[0x000001] : %s %s doesn't exists or not designated" % (
                name, getattr(options, name)))
            sys.exit(1)

    if options.script_dir and not os.path.isdir(options.script_dir):
        sys.stderr.write(
            "[0x000001] : previous script directory  %s doesn't exists or not designated" % options.script_dir)
        sys.exit(1)
    logging.info(
        "==============================START SCRIPT GENERATIONS=========================================")

    logging.info("option is: \nyang_directory:%s\n"
                 "xml_directory:%s\n"
                 "log_directory:%s",
                 os.path.abspath(options.yang_dir),
                 os.path.abspath(options.xml_dir),
                 os.path.abspath(options.log_dir))
    return options


def get_number_of_para_xml(xml_dir):
    """Get the number of "xxx_full.xml" under xml_dir path(recursive function).
    Args:
        xml_dir: Parser form command line -r.
    Returns:
        full_number: The number of "xxx_full.xml" under xml_dir path.
    """
    full_number = 0
    xml_files = os.listdir(xml_dir)
    for xml_file in xml_files:
        path = os.path.join(xml_dir, xml_file)
        if os.path.isdir(path):
            full_number += get_number_of_para_xml(path)
        else:
            if os.path.isfile(path) and not xml_file.endswith("_example.xml"):
                full_number = full_number + 1
    return full_number


def generate_netconf_script(xml_dir, output_dir, script_dir, direct_sub, completed_progress,
                            average_remaining_progress):
    """Generate netconf script(recursive function).
    Args:
        xml_dir: Parser form command line -r.
        output_dir: Script output path.
        script_dir: Parser form command line -p.
        direct_sub: Indicates whether it is a direct sub directory.
        completed_progress: Record the completed progress.
        average_remaining_progress: Add progress.
    Returns:
        completed_progress: Record the completed progress.
    Raises:
        Exception: Capture execution exception.
    """
    global GENERATE_SCRIPT_MESSAGE
    try:
        xml_files = os.listdir(xml_dir)
        for xml_file in xml_files:
            path = os.path.join(xml_dir, xml_file)
            if os.path.isdir(path):
                sub_xml_dir = os.path.join(xml_dir, xml_file)
                if direct_sub:
                    sub_output_dir = os.path.join(output_dir, xml_file)
                    if os.path.exists(sub_output_dir):
                        shutil.rmtree(sub_output_dir)
                        logging.warning(
                            "sub directory %s removed", sub_output_dir)
                    os.makedirs(sub_output_dir)
                    if script_dir:
                        script_dir = os.path.join(SCRIPT_DIR_GLOBAL, xml_file)
                    if script_dir and not os.path.isdir(script_dir):
                        logging.warning(
                            "no corresponding direcotory %s ", script_dir)
                else:
                    sub_output_dir = output_dir
                completed_progress = generate_netconf_script(sub_xml_dir, sub_output_dir, script_dir, False,
                                                                 completed_progress, average_remaining_progress)
            else:
                script_name = os.path.basename(xml_dir)
                if os.path.isfile(path) and xml_file.endswith(".xml") and not xml_file.endswith('_example.xml'):
                    logging.info("parse para file %s ", xml_file)
                    kwargs = var.get_params(
                        path, script_name, output_dir, script_dir)
                    if kwargs:
                        script_gen.Operation(**kwargs).run()
                    # progress_bar
                    completed_progress += average_remaining_progress
                    if completed_progress > 99:
                        completed_progress = 99
                    process_message = " {0} generated.".format(script_name)
                    base_util.print_progress_bar(completed_progress, process_message)
                    script_path = os.path.join(output_dir, script_name+'.py')
                    if os.path.exists(script_path):
                        GENERATE_SCRIPT_MESSAGE += "\nThe generated script has been saved to the path:{0}"\
                            .format(script_path)
                    else:
                        msg = "\n" + script_path + ". Cannot be generated."
                        base_util.error_write(msg)
    except Exception as error:
        logging.error("generate_netconf_script failed: %s",
                      traceback.format_exc())
    finally:
        return completed_progress


def main():
    """Main function.
    Raises:
        KeyboardInterrupt: Exit progress by keyboard:[ Ctrl + c ].
    """
    options = None
    try:
        options = env_parse()

        # Automatically deploy
        # if get_ansible_path():
        #     deploy(get_ansible_path()[1])
        xml_num = get_number_of_para_xml(options.xml_dir)
        if xml_num:
            get_parser.get_yang_handlers(options.yang_dir, options.xml_dir)
            # progress_bar
            completed_progress = 50
            process_message = " Yang Parser Completed."
            base_util.print_progress_bar(completed_progress, process_message)
            average_remaining_progress = int(math.floor((99 - completed_progress) / xml_num))
            if average_remaining_progress == 0:
                average_remaining_progress = 1
            output_dir = options.output_dir
            if output_dir is None:
                output_dir = get_ansible_path()[0]
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)
            if options.script_dir:
                global SCRIPT_DIR_GLOBAL
                SCRIPT_DIR_GLOBAL = options.script_dir
            generate_netconf_script(options.xml_dir, output_dir, options.script_dir, True, completed_progress,
                                    average_remaining_progress)
            logging.info(
                "==============================FINISH SCRIPT GENERATION==============================")
            process_message = " Finish Script Generation."
            base_util.print_progress_bar(100, process_message)
            sys.stdout.write(GENERATE_SCRIPT_MESSAGE+'\n')
            sys.stdout.flush()

    except KeyboardInterrupt:
        pass
    except Exception:
        logging.error("ANSIBLE GEN ERROR !")
        sys.stderr.write("[0x000000]Ansible Gen ERROR \n!")
        traceback.print_exc()
        logging.error(traceback.format_exc())
        sys.exit(1)
    finally:
        if options is not None:
            parse_error_and_exit(os.path.join(
                options.log_dir, SCRIPT_GEN_LOG_FILE))


if __name__ == "__main__":
    main()



import logging
import os
import traceback
import re
import sys
from pyang.context import Context
from pyang.repository import FileRepository
import pyang
if sys.version_info[0] == 2:
    import io
def init_ctx(path='./yang'):
    repos = FileRepository(path)
    ctx = Context(repos)
    return ctx


def parse_yang_module(yang_file_path, ctx):
    r = re.compile(r"^(.*?)(\@(\d{4}-\d{2}-\d{2}))?\.(yang|yin)$")
    r1 = re.compile(r"^(.*?)\.(yang|yin)$")
    try:
        if sys.version_info[0] == 2:
            fd = io.open(yang_file_path, "r", encoding="UTF-8")
        else:
            fd = open(yang_file_path, "r", encoding="UTF-8")
        text = fd.read()
        m1 = r1.search(yang_file_path)
        if m1 is None:
            return ctx
        m = r.search(yang_file_path)
        ctx.yin_module_map = {}
        if m is not None:
            (name, _dummy, rev, format) = m.groups()
            name = os.path.basename(name)
            ctx.add_module(yang_file_path, text, format, name, rev)
        else:
            ctx.add_module(yang_file_path, text)
    except Exception:
        yang_error_write("can not open the file %s" % yang_file_path)
    finally:
        if fd is not None:
            fd.close()
    logging.info('parse yang module %s success!', yang_file_path)
    return ctx


def parse_yang_modules(yang_directory, ctx):
    if os.path.isfile(yang_directory):
        parse_yang_module(yang_directory, ctx)
    else:
        for yang_file in sorted(os.listdir(yang_directory)):
            yang_file_path = os.path.join(yang_directory, yang_file)
            if os.path.isdir(yang_file_path):
                parse_yang_modules(yang_file_path, ctx)
            parse_yang_module(yang_file_path, ctx)
    ctx.validate()
    return ctx

def get_module_namespace(main_module):
    for stmt in main_module.substmts:
        if stmt.keyword == 'namespace':
            return stmt.arg
    return None
def is_filter(node):
    if not node.keyword == 'leaf':
        return True
    if hasattr(node,'i_is_key'):
        if node.i_is_key:
            return True
    for sub_stmt in node.substmts:
        if hasattr(sub_stmt,'i_extension'):
            if sub_stmt.keyword == ('huawei-extension', 'support-filter'):
                return True

    return False
def yang_module_has_error(ctx,module_name=None):
    for p,t,a in ctx.errors:
        if module_name is None:
            if is_error(t):
                return True
        else:
            if (is_error(t) and check_error_if_need(p, module_name)):
                return True

    return False

def check_error_if_need(pos, module_name):
    yang_error_name = ''
    try:
        yang_error_name = pos.top.arg
    except:
        pass
    if yang_error_name == module_name:
        return True
    return False

def is_error(error_type):
    error_level = pyang.error.err_level(error_type)
    if pyang.error.is_error(error_level):
        return True
    else:
        return False
def print_yang_errors(ctx, module_name=None):
    for p, t, a in ctx.errors:
        error_str = None
        error_level = pyang.error.err_level(t)
        if pyang.error.is_error(error_level):
            error_level_str = "Error"
        else:
            error_level_str = "Warning"
        if module_name is not None:
            if is_error(t) and check_error_if_need(p, module_name):
                error_str = ''.join([p.label(), " ", error_level_str,': ',pyang.error.err_to_str(t, a)])
        else:
            error_str = ''.join([p.label(), " ", error_level_str,': ',pyang.error.err_to_str(t, a)])
        if error_str is not None:
            print(error_str)
            yang_error_write(error_str)
    return

def yang_error_write(error_string):
    exc_info = sys.exc_info()
    exc_value = exc_info[1]

    if exc_value and traceback.format_exc().strip() != 'None':
        logging.error(traceback.format_exc())
    logging.error("%s",error_string)


from setuptools import setup, find_packages

setup(
    name='ansible-gen',
    version='0.5.2',
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
          'ansible-gen=ansible_gen.ansible_gen:main',
        ],
    },
    install_requires=[
        "pyang >=2.5.0",
        "jinja2==2.11.0",
        "xmltodict==0.11.0",
    ],
    url='https://github.com/HuaweiDatacomm/ansible-gen',
    license='Apache License, Version 2.0',
    author='frank feng',
    author_email='frank.fengchong@huawei.com',
    description='a tool can generate ansible api according yang modules and some input xml files'
)



#!/usr/bin/env python
# -*- coding: utf-8


import re
import logging
from xml.dom.minidom import parse
from xml.parsers.expat import ExpatError
from collections import OrderedDict
import xmltodict
from ..base_util import error_write, xml_structure_except

CONFIG_OR_FILTER_CONTENT = OrderedDict()
XMLNS_INFO_LIST = []

def process_xml_dict(xml_dict,pkg_type):
    rpc_xml_dict = xml_dict['rpc']
    if pkg_type == 'config':
        edit_config_xml_dict = rpc_xml_dict['edit-config']
        return edit_config_xml_dict['config']
    elif pkg_type in ['get','get-config']:
        get_xml_dict = rpc_xml_dict[pkg_type]
        return  get_xml_dict['filter']
    else:
        return rpc_xml_dict

def get_config_or_filter(xml_dic):
    """Get filter/config's descendant node.(recursive function).
    Args:
        xml_dic: An OrderedDict. Mapping xml's structure.
    Returns:
        CONFIG_OR_FILTER_CONTENT: global var.Used to save [config] or [filter] content.
        for example:
            OrderedDict([('if:interfaces',
                OrderedDict([('@xmlns:if', 'urn:ietf:params:xml:ns:yang:ietf-interfaces'),
                             ('@xmlns:ni', 'urn:ietf:params:xml:ns:yang:ietf-network-instance'), ('if:interface',
                    OrderedDict([('if:name', None), ('ip:ipv4',
                        OrderedDict([('@xmlns:ip', 'urn:ietf:params:xml:ns:yang:ietf-ip'), ('ip:address',
                            OrderedDict([('ip:prefix-length', '24'), ('ip:ip', '10.1.1.2')]))]))]))]))])
    """
    for item in xml_dic:
        if item == 'config' or item == 'filter':
            global CONFIG_OR_FILTER_CONTENT
            CONFIG_OR_FILTER_CONTENT = xml_dic[item]
        else:
            if item.find("@") < 0 and item.find("#") < 0:
                # recursive when the value is instance of OrderedDict.
                if isinstance(xml_dic[item], OrderedDict):
                    get_config_or_filter(xml_dic[item])

def get_rpc_input(xml_dic):
    for item in xml_dic["rpc"]:
        if item.find("@") < 0 and item.find("#") < 0:
            global CONFIG_OR_FILTER_CONTENT
            if(xml_dic["rpc"] is None):
                print(xml_dic["rpc"])
            if CONFIG_OR_FILTER_CONTENT is None:
                print(CONFIG_OR_FILTER_CONTENT)
            CONFIG_OR_FILTER_CONTENT[item] = xml_dic["rpc"][item]
# def generate_xmlns_info(parent_xpath,parent_xmlns,xmlns_list,current_node_content):
#     if current_node_content is None or len(current_node_content) == 0:
#         return
#     for current_node_name in current_node_content:
#         # Whether is element node
#         if current_node_name.find("@") >= 0 or current_node_name.find("#") >= 0:
#             continue
#         xpath = parent_xpath + '/' + current_node_name

def generate_xmlns_info_with_recursive(current_node_content, xpath):
    """Generate xmlns_info with recursive.(recursive function).
    Args:
        current_node_content: An OrderedDict. Used to record current node's content.
        xpath: Used to record the node's xpath.
    Returns:
        XMLNS_INFO_LIST: global list.Recording all nodes's xmlns information.
        structure like:
            [
             { '/node1/prefix2:node2':['node2-value','xmlns2'] },
             { '/node1':['node1-value','xmlns1'] },
             { '/node1/prefix2:node2/prefix3:node3':['node3-value','xmlns3'] },
             ...
            ]
    """
    for current_node_name in current_node_content:
        # Whether is element node
        if current_node_name.find("@") < 0 and current_node_name.find("#") < 0:
            xpath = xpath + '/' + current_node_name
            xmlns_str = ''
            text_str = ''
            global XMLNS_INFO_LIST
            # case: ( 'currentElementNodeName',OrderedDict([(...)]) ):
            if isinstance(current_node_content[current_node_name], OrderedDict):
                for child_node_name in current_node_content[current_node_name]:
                    if child_node_name.find("@xmlns") >= 0:
                        xmlns_str = xmlns_str + child_node_name + '="' + \
                                    current_node_content[current_node_name][child_node_name] + '"'
                        continue
                    if child_node_name.find("#text") >= 0:
                        text_str = current_node_content[current_node_name][child_node_name]
                        continue
                    if child_node_name.find("@") < 0 and child_node_name.find("#") < 0:
                        child_xpath = xpath + '/' + child_node_name
                        # case: ( 'childElementNodeName',OrderedDict([(...)]) ):
                        if isinstance(current_node_content[current_node_name][child_node_name], OrderedDict):
                            # save child_xpath info before recursive.
                            child_xmlns_str = ''
                            child_text_str = ''
                            child_xpath_dict = dict()
                            # case: OrderedDict([('@xmlns:prefix1','http:xxxx'),('@xmlns:prefix2','urn:xxxx'),(...)])
                            for grandson_node_name in current_node_content[current_node_name][child_node_name]:
                                if grandson_node_name.find("@xmlns") >= 0:
                                    child_xmlns_str = child_xmlns_str + grandson_node_name + '="' + \
                                                      current_node_content[current_node_name][child_node_name][grandson_node_name] + '"'
                                    continue
                            # case: OrderedDict([(...),('#text','childTextContent')])
                            for grandson_node_name in current_node_content[current_node_name][child_node_name]:
                                if grandson_node_name.find("#text") >= 0:
                                    child_text_str = current_node_content[current_node_name][child_node_name][grandson_node_name]
                                    continue
                            child_xpath_dict[child_xpath] = [child_text_str, child_xmlns_str]
                            XMLNS_INFO_LIST.append(child_xpath_dict)
                            # recursive
                            generate_xmlns_info_with_recursive(current_node_content[current_node_name][child_node_name], child_xpath)
                        # case: ( 'childElementNodeName','childTextContent' )
                        elif isinstance(current_node_content[current_node_name][child_node_name], str):
                            child_text_str = current_node_content[current_node_name][child_node_name]
                            child_xpath_dict = dict()
                            if child_text_str:
                                child_xpath_dict[child_xpath] = [child_text_str, '']
                            else:
                                child_xpath_dict[child_xpath] = ['', '']
                            XMLNS_INFO_LIST.append(child_xpath_dict)
                        # case: ( 'childElementNodeName',None )
                        # current_node_content[current_node_name][child_node_name] is None:
                        else:
                            child_xpath_dict = dict()
                            child_xpath_dict[child_xpath] = ['', '']
                            XMLNS_INFO_LIST.append(child_xpath_dict)
            # case: ( 'currentElementNodeName','currentTextContent' ):
            else:
                text_str = current_node_content[current_node_name]
            if text_str is None:
                text_str = ''
            xpath_dict = dict()
            xpath_dict[xpath] = [text_str, xmlns_str]
            XMLNS_INFO_LIST.append(xpath_dict)
            xpath = xpath[:-(len(current_node_name)+1)]

def generate_xmlns_info(file,pkg_type):
    global XMLNS_INFO_LIST
    XMLNS_INFO_LIST = []
    try:
        with open(file, 'r') as read:
            xml_str = read.read()
            xml_dic = xmltodict.parse(xml_str)
            xml_dic = process_xml_dict(xml_dic,pkg_type)
            if xml_dic is None:
                return None
            xpath = ''
            generate_xmlns_info_with_recursive(xml_dic, xpath)
    except ExpatError as expat_exception:
        xml_structure_except(expat_exception, file)
    except Exception as error_str:
        error_write(error_str)
    finally:
        return XMLNS_INFO_LIST


# def generate_xmlns_info(file):
#     """Generate xmlns_info.
#     Args:
#         file: The path of xml file.
#     Returns:
#         XMLNS_INFO_LIST: global list.Recording all nodes's xmlns information.
#         structure like:
#             [
#              { '/node1/prefix2:node2':['node2-value','xmlns2'] },
#              { '/node1':['node1-value','xmlns1'] },
#              { '/node1/prefix2:node2/prefix3:node3':['node3-value','xmlns3'] },
#              ...
#             ]
#     Raises:
#         Exception: Capture execution exception.
#     """
#     global CONFIG_OR_FILTER_CONTENT
#     global XMLNS_INFO_LIST
#     CONFIG_OR_FILTER_CONTENT = OrderedDict()
#     XMLNS_INFO_LIST = []
#
#     try:
#         with open(file, 'r') as read:
#             xml_str = read.read()
#             xml_dic = xmltodict.parse(xml_str)
#             get_config_or_filter(xml_dic)
#             if not CONFIG_OR_FILTER_CONTENT:
#                 get_rpc_input(xml_dic)
#             xpath = ''
#             config_or_filter_content = CONFIG_OR_FILTER_CONTENT
#             if config_or_filter_content:
#                 generate_xmlns_info_with_recursive(config_or_filter_content, xpath)
#             else:
#                 logging.info("There is no <config> or <filter>, please check:" + file)
#     except ExpatError as expat_exception:
#         xml_structure_except(expat_exception, file)
#     except Exception as error_str:
#         error_write(error_str)
#     finally:
#         return XMLNS_INFO_LIST

def order_by_xpath(elem):
    """
    Sort the elem by length of xpathNode.
    """
    for key in elem:
        xpath_result = key.split("/")
    return len(xpath_result)


def get_xmlns_info(file,pkg_type):
    """Get sorted xmlns_info.
    Args:
        file: The path of xml file.
    Returns:
        sort_result: A list.Recording all nodes's xmlns information.
        structure like:
            [
             { '/node1':['node1-value','xmlns1'] },
             { '/node1/prefix2:node2':['node2-value','xmlns2'] },
             { '/node1/prefix2:node2/prefix3:node3':['node3-value','xmlns3'] },
             ...
            ]
    """
    sort_result = generate_xmlns_info(file,pkg_type)
    sort_result.sort(key=order_by_xpath)
    return sort_result


def xmlns_info_without_prefix(xmlns_info):
    """Remove prefix from xpath-key.
    Args:
        xmlns_info: The xmlns_info.
    Returns:
        new_xmlns_info_list: In xmlns_info,the xpath-key without prefix.
        for example:
            xmlns_info:
                [ { '/node1/prefix2:node2/prefix3:node3':[ 'node3-value','xmlns3' ] },... ]
            new_xmlns_info_list:
                [ { '/node1/node2/node3':[ 'node3-value','xmlns3','/node1/prefix2:node2/prefix3:node3' ] },... ]
    """
    new_xmlns_info_list = []
    for item in xmlns_info:
        new_dict = {}
        for xapth_key in item:
            xapth_key_list = xapth_key.split("/")[1:]
            new_xpath_key = ''
            for each_node in xapth_key_list:
                new_xpath_key = new_xpath_key + "/" + each_node.split(":")[-1]
            temp_list = item[xapth_key]
            temp_list.append(xapth_key)
        new_dict[new_xpath_key] = temp_list
        new_xmlns_info_list.append(new_dict)
    return new_xmlns_info_list


def get_xml_namespace(file_name,pkg_type):
    """Get xml's namespace.
    Args:
        file_name: The path of xml file.
    Returns:
        xml_namespace: The namespace of xml.
        for example:
            xml file content:
                ...
                <config xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
                    <interfaces xmlns="urn:ietf:params:xml:ns:yang:ietf-interfaces"
                                xmlns:ni="urn:ietf:params:xml:ns:yang:ietf-network-instance">
                        ...
                    </interfaces>
                </config>
            xml_namespace: 'urn:ietf:params:xml:ns:yang:ietf-interfaces'
    Raises:
        Exception: Capture execution exception.
    """
    feature_namespaces = []
    try:
        doc = parse(file_name)
        root = doc.documentElement
        child_nodes = None
        if pkg_type in ['get','get-config']:
            if root.getElementsByTagNameNS("urn:ietf:params:xml:ns:netconf:base:1.0", "filter"):
                child_nodes = root.getElementsByTagNameNS("urn:ietf:params:xml:ns:netconf:base:1.0", "filter")[
                    0].childNodes
        elif pkg_type == 'config':
            if root.getElementsByTagNameNS("urn:ietf:params:xml:ns:netconf:base:1.0", "config"):
                child_nodes = root.getElementsByTagNameNS("urn:ietf:params:xml:ns:netconf:base:1.0", "config")[
                    0].childNodes
        else:
            child_nodes = root.childNodes
            logging.info("This is rpc-xml:" + file_name)

        if child_nodes is not None:
            for child_node in child_nodes:
                if child_node.nodeType == 1 and hasattr(child_node, 'namespaceURI'):
                    feature_namespaces.append(child_node.namespaceURI)

    except ExpatError as expat_exception:
        xml_structure_except(expat_exception, file_name)
    except Exception as error_str:
        error_write(error_str)
    return feature_namespaces


def get_node_xmlns(xpath, xmlns_info, pre_element_node_prefix=""):
    """According the xmlns_info to query node's namespace.(recursive function).
    Args:
        xpath: The xpath node which need to find namespace.(without prefix).
        xmlns_info: The xmlns_info.
            structure like:
                [ { '/node1/node2/node3':[ 'node3-value','xmlns3','/node1/prefix2:node2/prefix3:node3' ] },... ].
        pre_element_node_prefix: Used to record the parent node prefix.
    Returns:
        The node's xmlns.
        for example:
            xpath: '/interfaces/interface'
            xmlns_info:
                [{'/interfaces': ['', '@xmlns:if="urn:ietf:params:xml:ns:yang:ietf-interfaces"', '/if:interfaces']},
                 {'/interfaces/interface': ['', '', '/if:interfaces/if:interface']},
                 ...]
            return: 'urn:ietf:params:xml:ns:yang:ietf-interfaces'
    """
    if xpath == '':
        return 'urn:ietf:params:xml:ns:netconf:base:1.0'
    for item in xmlns_info:
        if xpath in item.keys():
            xmlns = item[xpath][1]

            xpath_with_prefix = item[xpath][2]
            end_element_node_with_prefix = xpath_with_prefix.split("/")[-1]
            end_element_node_list = end_element_node_with_prefix.split(":")
            if len(end_element_node_list) == 1:
                element_node_prefix = ""
            elif len(end_element_node_list) == 2:
                element_node_prefix = end_element_node_list[0]
            else:
                logging.info("the len of element_node_prefix not equal 1 or 2,please check xpath: " +
                             str(xpath))

            if pre_element_node_prefix != "" and pre_element_node_prefix != element_node_prefix:
                element_node_prefix = pre_element_node_prefix
            if xmlns:
                if element_node_prefix == "":
                    search_obj_dual = re.search(r'xmlns="(.*?)"', xmlns)
                    search_obj_single = re.search(r"xmlns='(.*?)'", xmlns)
                    if search_obj_dual:
                        return search_obj_dual.group(1)
                    elif search_obj_single:
                        return search_obj_single.group(1)
                    else:
                        parent_node_xpath = xpath[:-(len(xpath.split("/")[-1])+1)]
                        return get_node_xmlns(parent_node_xpath, xmlns_info)
                else:
                    search_obj_dual = re.search(r'xmlns:' + element_node_prefix + '="(.*?)"', xmlns)
                    search_obj_single = re.search(r"xmlns:" + element_node_prefix + "='(.*?)'", xmlns)
                    if search_obj_dual:
                        return search_obj_dual.group(1)
                    if search_obj_single:
                        return search_obj_single.group(1)
            else:
                if element_node_prefix == "":
                    parent_node_xpath = xpath[:-(len(xpath.split("/")[-1])+1)]
                    return get_node_xmlns(parent_node_xpath, xmlns_info)
            parent_node_xpath = xpath[:-(len(xpath.split("/")[-1])+1)]
            return get_node_xmlns(parent_node_xpath, xmlns_info, element_node_prefix)
